import sys

import matplotlib
import matplotlib.style as mplstyle
mplstyle.use('fast')

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D

from CADETMatch.cache import cache

from pathlib import Path
import pandas
import numpy
import scipy.interpolate
import itertools

from cadet import H5
from addict import Dict

# parallelization
import multiprocessing

import os
import warnings
import corner
import seaborn as sns
import CADETMatch.util as util
import logging
import CADETMatch.loggerwriter as loggerwriter

def new_range(flat_chain):
    lb_data, mid_data, ub_data = numpy.percentile(flat_chain, [5, 50, 95], 0)

    distance = numpy.max(numpy.abs(numpy.array([lb_data - mid_data, ub_data - mid_data])), axis=0)

    lb_range = mid_data - 2 * distance
    ub_range = mid_data + 2 * distance

    return numpy.array([lb_range, ub_range])

def main():
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "graph_corner.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("graphing directory %s", os.getcwd())

    graphCorner(cache)


def plotChain(flat_chain, flat_chain_transform, headers, out_dir, prefix):
    multiprocessing.get_logger().info("plotting chain")
    if len(flat_chain) > 1e5:
        indexes = numpy.random.choice(flat_chain.shape[0], int(1e5), replace=False)
        chain = flat_chain[indexes]
        chain_transform = flat_chain_transform[indexes]
    else:
        chain = flat_chain
        chain_transform = flat_chain_transform

    if len(chain) == 0:
        return

    # stat corner plot
    fig_size = 6 * len(headers)
    fig = corner.corner(
        chain,
        quantiles=(0.16, 0.5, 0.84),
        show_titles=True,
        labels=headers,
        bins=20,
        range=new_range(chain).T,
        use_math_text=True,
        title_fmt=".2g",
    )
    fig.set_size_inches((fig_size, fig_size))
    fig.savefig(str(out_dir / ("%s_corner.png" % prefix)))

    fig = corner.corner(
        chain_transform,
        quantiles=(0.16, 0.5, 0.84),
        show_titles=True,
        labels=headers,
        bins=20,
        range=new_range(chain_transform).T,
        use_math_text=True,
        title_fmt=".2g",
    )
    fig.set_size_inches((fig_size, fig_size))
    fig.savefig(str(out_dir / ("%s_corner_transform.png" % prefix)))


def plotMCMCParam(out_dir, param, chain, header):
    multiprocessing.get_logger().info("plotting mcmc param %s", param)
    shape = chain.shape
    x = numpy.linspace(0, shape[1], shape[1])

    fig = figure.Figure(figsize=[10, 10])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    graph.plot(x, chain[0, :, param], "r", label="5")
    graph.plot(x, chain[1, :, param], "k", label="mean")
    graph.plot(x, chain[2, :, param], "r", label="95")
    graph.fill_between(x, chain[0, :, param], chain[2, :, param], facecolor="r", alpha=0.5)
    graph.legend()
    graph.set_title(header)
    fig.set_size_inches((12, 12))
    print(header, header.replace('/', '_'))
    fig.savefig(str(out_dir / ("%s.png" % header.replace('/', '_'))))


def plotMCMCVars(out_dir, headers, data):
    for param_idx, header in enumerate(headers):
        if "train_chain_stat" in data.root:
            plotMCMCParam(out_dir, param_idx, data.root.train_chain_stat, "Train " + header + str(param_idx))
            plotMCMCParam(out_dir, param_idx, data.root.train_chain_stat_transform, "Train " + header + " Transform" + str(param_idx))

        if "run_chain_stat" in data.root:
            plotMCMCParam(out_dir, param_idx, data.root.run_chain_stat, "Run " + header + str(param_idx))
            plotMCMCParam(out_dir, param_idx, data.root.run_chain_stat_transform, "Run " + header + " Transform" + str(param_idx))


def keep_header(char):
    allowed = set("-0123456789")
    return char in allowed


def clean_header(header):
    splits = header.split()
    temp = [splits[0]]
    for string in splits[1:]:
        temp.append("".join([i for i in string if keep_header(i)]))
    return " ".join(temp)


def graphCorner(cache):
    multiprocessing.get_logger().info("plotting corner plots")
    headers = list(cache.parameter_headers_actual)
    headers = [clean_header(header) for header in headers]

    resultDir = Path(cache.settings["resultsDir"])
    result_h5 = resultDir / "result.h5"

    miscDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = miscDir / "mcmc.h5"

    if mcmc_h5.exists():
        data = H5()
        data.filename = mcmc_h5.as_posix()
        data.load()

        out_dir = cache.settings["resultsDirProgress"]

        if "tau_percent" in data.root:
            multiprocessing.get_logger().info("plotting tau plots")
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)

            tau_percent = numpy.squeeze(data.root.tau_percent)
            max_values = numpy.ones(tau_percent.shape)
            tau_percent = numpy.min([tau_percent, max_values], 0)

            graph.bar(list(range(len(headers))), tau_percent, tick_label=headers)
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "tau_percent.png"))

        if "flat_chain" in data.root:
            multiprocessing.get_logger().info("plot flat chain")
            plotChain(data.root.flat_chain, data.root.flat_chain_transform, headers, out_dir, "full")

        if "burn_in_acceptance" in data.root:
            multiprocessing.get_logger().info("plot burn in acceptance")
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            graph.plot(util.get_confidence(data.root.burn_in_acceptance))
            graph.set_title("Burn In Acceptance")
            graph.set_xlabel("Step")
            graph.set_ylabel("Acceptance")
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "burn_in_acceptance.png"))

        if "integrated_autocorrelation_time" in data.root:
            iat = data.root.integrated_autocorrelation_time

            multiprocessing.get_logger().info("plot integrated autocorrelation time")
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)

            for i in range(1, iat.shape[1]):
                graph.plot(iat[:, 0], iat[:, i], label=i)
            graph.set_title("Integrated Autocorrelation Time")
            graph.set_xlabel("Step")
            graph.set_ylabel("IAT")
            graph.legend()
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "integrated_autocorrelation_time.png"))

            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            tau = cache.MCMCTauMult
            for i in range(1, iat.shape[1]):
                progress = (iat[:, 0] / tau) / iat[:, i]
                progress[progress > 1] = 1
                graph.plot(iat[:, 0], progress, label=i)
            graph.set_title("Integrated Autocorrelation Time Progress")
            graph.set_xlabel("Step")
            graph.set_ylabel("Progress")
            graph.legend()
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "integrated_autocorrelation_time_progress.png"))

        if "mcmc_acceptance" in data.root:
            multiprocessing.get_logger().info("plot mcmc acceptance")
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            graph.plot(util.get_confidence(data.root.mcmc_acceptance))
            graph.set_title("MCMC Acceptance")
            graph.set_xlabel("Step")
            graph.set_ylabel("Acceptance")
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "mcmc_acceptance.png"))

        plotChain(data.root.train_flat_chain, data.root.train_flat_chain_transform, headers, out_dir, "train")
        plotMCMCVars(out_dir, headers, data)

    else:
        data = H5()
        data.filename = result_h5.as_posix()
        data.load()

        if len(data.root.input) > 1e5:
            indexes = numpy.random.choice(data.root.input.shape[0], int(1e5), replace=False)
            data_input = data.root.input[indexes]
            data_input_transform = data.root.input_transform[indexes]
            weight_min = data.root.output_meta[indexes, 1]
            weight_prod = data.root.output_meta[indexes, 2]
            all_scores = data.root.output[indexes]
        else:
            data_input = data.root.input
            data_input_transform = data.root.input_transform
            weight_min = data.root.output_meta[:, 1]
            weight_prod = data.root.output_meta[:, 2]
            all_scores = data.root.output

        max_scores = numpy.max(all_scores, 1)
        acceptable = max_scores > 0.01

        if numpy.any(acceptable):
            multiprocessing.get_logger().info("graphing remove %s points", len(max_scores) - numpy.sum(acceptable))

            data_input = data_input[acceptable]
            data_input_transform = data_input_transform[acceptable]
            weight_min = weight_min[acceptable]
            weight_prod = weight_prod[acceptable]

        out_dir = cache.settings["resultsDirProgress"]

        create_corner(out_dir, "corner.png", headers, data_input, weights=None)
        create_corner(out_dir, "corner_min.png", headers, data_input, weights=weight_min)
        create_corner(out_dir, "corner_prod.png", headers, data_input, weights=weight_prod)

        # transformed entries
        create_corner(out_dir, "corner_transform.png", headers, data_input_transform, weights=None)
        create_corner(out_dir, "corner_min_transform.png", headers, data_input_transform, weights=weight_min)
        create_corner(out_dir, "corner_prod_transform.png", headers, data_input_transform, weights=weight_prod)


def create_corner(dir, filename, headers, data, weights=None):
    if len(data) <= len(headers):
        multiprocessing.get_logger().info("There is not enough data to generate corner plots")
        return
    if numpy.all(numpy.min(data, 0) < numpy.max(data, 0)):
        if weights is None or numpy.max(weights) > numpy.min(weights):
            bounds = new_range(data).T

            fig_size = 6 * len(headers)
            fig = corner.corner(
                data,
                quantiles=(0.16, 0.5, 0.84),
                show_titles=True,
                labels=headers,
                bins=20,
                range=new_range(data).T,
                use_math_text=True,
                title_fmt=".2g",
            )
            fig.set_size_inches((fig_size, fig_size))
            fig.savefig(str(dir / filename))

            df = pandas.DataFrame(data, columns=headers)
            fig = sns.PairGrid(df, diag_sharey=False, height=fig_size)
            fig.map_lower(sns.kdeplot)
            fig.map_upper(sns.scatterplot)
            fig.map_diag(sns.kdeplot, lw=3)

            for row_num, row in enumerate(bounds):
                for col_num, col in enumerate(bounds):
                    if row_num == col_num:
                        fig.axes[row_num, col_num].set_xlim(*row)
                    else:
                        fig.axes[row_num, col_num].set_xlim(*col)
                        fig.axes[row_num, col_num].set_ylim(*row)

            fig.savefig(str(dir / ("sns_%s" % filename)))


if __name__ == "__main__":
    main()



from scipy.optimize._numdiff import approx_derivative
from CADETMatch.cache import cache
from pathlib import Path
import CADETMatch.util as util
import numpy
import multiprocessing


class GradientException(Exception):
    pass


def jac(individual, cache):
    J = approx_derivative(fitness_grad, individual, method="3-point", kwargs={"cache": cache})
    cond = numpy.linalg.cond(J)
    u, s, v = numpy.linalg.svd(J)
    minSing = numpy.min(s)
    maxSing = numpy.max(s)
    multiprocessing.get_logger().info("%s has condition %s  min sing:  %s   max sing: %s", individual, cond, minSing, maxSing)
    return J


def fitness_grad(individual, cache):
    scores = []
    error = 0.0

    results = {}
    for experiment in cache.settings["experiments"]:
        result = runExperiment(individual, experiment, cache.settings, cache.target, cache)
        if result is not None:
            results[experiment["name"]] = result
            scores.extend(results[experiment["name"]]["scores"])
            error += results[experiment["name"]]["error"]
        else:
            raise GradientException("Gradient caused simulation failure, aborting")

    return scores


def runExperiment(individual, experiment, settings, target, cache):
    if "simulation" not in experiment:
        templatePath = Path(settings["resultsDirMisc"], "template_%s.h5" % experiment["name"])
        templateSim = Cadet()
        templateSim.filename = templatePath.as_posix()
        templateSim.load()
        experiment["simulation"] = templateSim

    return util.runExperiment(
        individual, experiment, settings, target, experiment["simulation"], experiment["simulation"].root.timeout, cache
    )



__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]


__title__ = "CADETMatch"
__summary__ = "CADETMatch is a parameter estimation and error modeling library for CADET"
__author__ = "William Heymann"
__email__ = "w.heymann@fz-juelich.de"
__license__ = "GNU General Public License v3 (GPLv3)"
__copyright__ = "2020 %s" % __author__
__version__ = "0.6.23"
__uri__ = "https://github.com/modsim/CADET-Match"



from deap import tools
from operator import eq
import math
import numpy
import multiprocessing
import copy

smallest = numpy.finfo(1.0).tiny

diff_step = 1e-1


class ParetoFront(tools.ParetoFront):
    "Modification of the pareto front in DEAP that takes cache as an argument to update to use for similar comparison"

    def __init__(self, similar=None, similar_fit=None, slice_object=None):
        if similar is None:
            similar = eq
        if similar_fit is not None:
            self.similar_fit = similar_fit
        else:
            self.similar_fit = eq
        self.slice_object = slice_object
        super().__init__(similar)

    def update(self, population, numGoals):
        """Update the Pareto front hall of fame with the *population* by adding 
        the individuals from the population that are not dominated by the hall
        of fame. If any individual in the hall of fame is dominated it is
        removed.
        
        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        new_members = []
        significant = []
        slice_object = self.slice_object
        pareto_length = len(self)
        if slice_object is None:
            slice_object = slice(None)

        for ind in population:
            is_dominated = False
            dominates_one = False
            has_twin = False
            to_remove = []
            for i, hofer in enumerate(self):  # hofer = hall of famer
                if not dominates_one and hofer.fitness.dominates(ind.fitness, obj=slice_object):
                    is_dominated = True
                    break
                elif ind.fitness.dominates(hofer.fitness, obj=slice_object):
                    dominates_one = True
                    to_remove.append(i)
                    significant.append(not self.similar_fit(ind.fitness.values, hofer.fitness.values))
                elif self.similar_fit(ind.fitness.values, hofer.fitness.values) and self.similar(ind, hofer):
                    has_twin = True
                    break

            for i in reversed(to_remove):  # Remove the dominated hofer
                self.remove(i)
            if not is_dominated and not has_twin:
                # if the pareto front is empty or a new item is added to the pareto front that is progress
                # however if there is only a single objective and it does not significantly dominate then
                # don't count that as significant progress
                if pareto_length == 0:
                    significant.append(True)
                elif numGoals > 1:
                    if len(significant) == 0 or (len(significant) and any(significant)):
                        significant.append(True)

                self.insert(ind)
                new_members.append(ind)

        return new_members, any(significant)

    def getBestScores(self):
        weights = numpy.array(self[0].fitness.weights)
        data_meta = numpy.array([i.fitness.values for i in self])
        return numpy.max(data_meta * weights, 0) * weights


class ParetoFrontMeta(ParetoFront):
    def __init__(self, similar=None, similar_fit=None, slice_object=None):
        super().__init__(similar, similar_fit, slice_object)
        self.best_sse = numpy.inf
        self.best_sse_ind = None
        self.best_rmse = numpy.inf
        self.best_rmse_end = None

    def update(self, population, numGoals):
        """Update the Pareto front hall of fame with the *population* by adding 
        the individuals from the population that are not dominated by the hall
        of fame. If any individual in the hall of fame is dominated it is
        removed.
        
        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        new_members = []
        significant = []
        slice_object = self.slice_object
        pareto_length = len(self)
        if slice_object is None:
            slice_object = slice(None)

        for ind in population:
            is_dominated = False
            dominates_one = False
            has_twin = False
            to_remove = []

            ind_sse = ind.fitness.values[-2]
            ind_rmse = ind.fitness.values[-1]

            new_best_sse_found = False
            new_best_rmse_found = False

            if ind_sse < self.best_sse:
                new_best_sse = ind_sse
                new_best_sse_ind = copy.deepcopy(ind)
                new_best_sse_ind.best = True
                new_best_sse_found = True

            if ind_rmse < self.best_rmse:
                new_best_rmse = ind_rmse
                new_best_rmse_ind = copy.deepcopy(ind)
                new_best_rmse_ind.best = True
                new_best_rmse_found = True

            for i, hofer in enumerate(self):  # hofer = hall of famer
                if not dominates_one and hofer.fitness.dominates(ind.fitness, obj=slice_object):
                    is_dominated = True
                    break
                elif ind.fitness.dominates(hofer.fitness, obj=slice_object):
                    dominates_one = True
                    to_remove.append(i)
                    significant.append(not self.similar_fit(ind.fitness.values, hofer.fitness.values))
                elif self.similar_fit(ind.fitness.values, hofer.fitness.values) and self.similar(ind, hofer):
                    has_twin = True
                    break

            for i in reversed(to_remove):  # Remove the dominated hofer
                self.remove(i)
            if not is_dominated and not has_twin:
                # if the pareto front is empty or a new item is added to the pareto front that is progress
                # however if there is only a single objective and it does not significantly dominate then
                # don't count that as significant progress
                if pareto_length == 0:
                    significant.append(True)
                elif numGoals > 1:
                    if len(significant) == 0 or (len(significant) and any(significant)):
                        significant.append(True)

                self.insert(ind)
                new_members.append(ind)

            # add sse and rmse individuals if they are not stored already
            foundSSE = False
            foundRMSE = False

            if new_best_rmse_found or new_best_sse_found:
                # remove old best
                to_remove = []
                for idx, ind in enumerate(self):
                    if ind.best:
                        removeSSE = new_best_sse_found and ind.fitness.values == self.best_sse_ind.fitness.values
                        removeRMSE = new_best_rmse_found and ind.fitness.values == self.best_rmse_ind.fitness.values
                        if removeSSE or removeRMSE:
                            to_remove.append(idx)

                for idx in reversed(to_remove):  # Remove the dominated hofer
                    self.remove(idx)

                if new_best_sse_found:
                    self.best_sse_ind = new_best_sse_ind
                    self.best_sse = new_best_sse

                if new_best_rmse_found:
                    self.best_rmse_ind = new_best_rmse_ind
                    self.best_rmse = new_best_rmse

            sameInd = self.best_sse_ind.fitness.values == self.best_rmse_ind.fitness.values

            for ind in self:
                if ind.fitness.values == self.best_sse_ind.fitness.values:
                    foundSSE = True
                if ind.fitness.values == self.best_rmse_ind.fitness.values:
                    foundRMSE = True
                if foundSSE and foundRMSE:
                    break
            if not foundSSE and sameInd:
                self.insert(self.best_sse_ind)
                new_members.append(ind)

            if not foundSSE and not sameInd:
                self.insert(self.best_sse_ind)
                new_members.append(ind)
            if not foundRMSE and not sameInd:
                self.insert(self.best_rmse_ind)
                new_members.append(ind)

        return new_members, any(significant)


class DummyFront(tools.ParetoFront):
    "Modification of the pareto front in DEAP that takes cache as an argument to update to use for similar comparison"

    def __init__(self, similar=None):
        "This is here for API compatibility, don't do anything"
        if similar is None:
            similar = eq
        super().__init__(similar)

    def update(self, population, live_mode=True):
        "do not put anything in this front, it is just needed to maintain compatibility"
        return [], False


def similar(a, b):
    "we only need a parameter to 4 digits of accuracy so have the pareto front only keep up to 5 digits for members of the front"
    a = numpy.absolute(numpy.array(a))
    b = numpy.absolute(numpy.array(b))

    # used to catch division by zero
    a[a == 0.0] = smallest

    diff = numpy.abs((a - b) / a)
    return numpy.all(diff < diff_step)


def similar_fit_norm(a, b):
    "we only need a parameter to 4 digits of accuracy so have the pareto front only keep up to 5 digits for members of the front"
    a = numpy.absolute(numpy.array(a))
    b = numpy.absolute(numpy.array(b))

    # used to catch division by zero
    a[a == 0.0] = smallest

    a = numpy.log(1 - a)
    b = numpy.log(1 - b)

    diff = numpy.abs((a - b) / a)
    return numpy.all(diff < diff_step)


def similar_fit_sse(a, b):
    "we only need a parameter to 4 digits of accuracy so have the pareto front only keep up to 5 digits for members of the front"
    a = numpy.absolute(numpy.array(a))
    b = numpy.absolute(numpy.array(b))

    # used to catch division by zero
    a[a == 0.0] = smallest

    a = numpy.log(a)
    b = numpy.log(b)

    diff = numpy.abs((a - b) / a)
    return numpy.all(diff < diff_step)


def similar_fit_meta_norm(a, b):
    a = numpy.array(a)
    b = numpy.array(b)

    # used to catch division by zero
    a[a == 0.0] = smallest

    # SSE is in the last slot so we only want to use the first 3 meta scores
    a = numpy.log(1 - a[:3])
    b = numpy.log(1 - b[:3])

    diff = numpy.abs((a - b) / a)
    return numpy.all(diff < diff_step)


def similar_fit_meta_sse(a, b):
    "SSE is negative and in the last slot and the only score needed"
    a = a[-2]
    b = b[-2]

    # used to catch division by zero
    if a == 0.0:
        a = smallest

    # SSE should be handled in log scale
    a = math.log(a)
    b = math.log(b)

    diff = numpy.abs((a - b) / a)
    return numpy.all(diff < diff_step)


def similar_fit_meta_sse_split(a, b):
    "SSE is negative and in the last slot and the only score needed"
    a = numpy.abs(numpy.array(a))
    b = numpy.abs(numpy.array(b))

    # used to catch division by zero
    a[a == 0.0] = smallest

    # SSE should be handled in log scale
    a = numpy.log(a[:3])
    b = numpy.log(b[:3])

    diff = numpy.abs((a - b) / a)
    return numpy.all(diff < diff_step)


def similar_fit_meta(cache):
    if cache.allScoreNorm:
        return similar_fit_meta_norm
    elif cache.allScoreSSE:
        if cache.MultiObjectiveSSE:
            return similar_fit_meta_sse_split
        else:
            return similar_fit_meta_sse


def similar_fit(cache):
    if cache.allScoreNorm:
        return similar_fit_norm
    elif cache.allScoreSSE:
        return similar_fit_sse


def updateParetoFront(halloffame, offspring, cache):
    new_members, significant = halloffame.update([offspring,], cache.numGoals)
    return bool(new_members), significant



import numpy
import CADETMatch.util as util

from pathlib import Path

from deap import creator
from deap import tools

from cadet import Cadet

import CADETMatch.cache as cache
import CADETMatch.score_calc as score_calc
import CADETMatch.progress as progress

import multiprocessing

ERROR = {"scores": None, "path": None, "simulation": None, "error": None, "cadetValues": None, "cadetValuesKEQ": None}


def fitness_final(individual, json_path, run_experiment=None):
    return fitness_base(runExperimentFinal, "simulation_final", individual, json_path, run_experiment)


def fitness(individual, json_path, run_experiment=None):
    return fitness_base(runExperiment, "simulation", individual, json_path, run_experiment)


def meta_score_trans(cache, score):
    temp = numpy.copy(score)

    if cache.allScoreSSE:
        temp[:3] = -temp[:3]
    else:
        temp[:3] = 1 - temp[:3]

    return temp


def fitness_base(runner, template_name, individual, json_path, run_experiment):
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings["resultsDirLog"], "main.log")
        cache.cache.setup(json_path)

    if run_experiment is None:
        run_experiment = runExperiment

    scores = []
    error = 0.0
    exp_values = []
    sim_values = []

    results = {}
    for experiment in cache.cache.settings["experiments"]:
        result = runner(individual, template_name, experiment, cache.cache.settings, cache.cache.target, cache.cache)
        if result is not None:
            results[experiment["name"]] = result
            scores.extend(results[experiment["name"]]["scores"])
            error += results[experiment["name"]]["error"]

            sim_values.extend(result["sim_value"])
            exp_values.extend(result["exp_value"])
        else:
            return cache.cache.WORST, [], cache.cache.WORST_META, None, individual

    rmse = score_calc.rmse_combine(exp_values, sim_values)

    if numpy.any(numpy.isnan(scores)):
        multiprocessing.get_logger().info("NaN found for %s %s", individual, scores)

    # human scores
    meta_score = numpy.concatenate([util.calcMetaScores(scores, cache.cache), [error, rmse]])

    for result in results.values():
        if result["cadetValuesKEQ"]:
            cadetValuesKEQ = result["cadetValuesKEQ"]
            break

    # generate csv
    csv_record = []
    csv_record.extend(["EVO", "NA"])
    csv_record.extend(cadetValuesKEQ)
    csv_record.extend(progress.score_trans(cache.cache, scores))
    csv_record.extend(meta_score_trans(cache.cache, meta_score))

    return scores, csv_record, meta_score, results, tuple(individual)


def saveExperiments(save_name_base, settings, target, results):
    return util.saveExperiments(save_name_base, settings, target, results, settings["resultsDirEvo"], "%s_%s_EVO.h5")


def plotExperiments(save_name_base, settings, target, results):
    util.plotExperiments(save_name_base, settings, target, results, settings["resultsDirEvo"], "%s_%s_EVO.png")


def runExperimentFinal(individual, template_name, experiment, settings, target, cache):
    sim_name = "template_%s_final.h5" % experiment["name"]
    return runExperimentBase(sim_name, template_name, individual, experiment, settings, target, cache)


def runExperiment(individual, template_name, experiment, settings, target, cache):
    sim_name = "template_%s.h5" % experiment["name"]
    return runExperimentBase(sim_name, template_name, individual, experiment, settings, target, cache)


def runExperimentBase(sim_name, template_name, individual, experiment, settings, target, cache):
    if template_name not in experiment:
        templatePath = Path(settings["resultsDirMisc"], sim_name)
        templateSim = Cadet()
        templateSim.filename = templatePath.as_posix()
        templateSim.load()
        experiment[template_name] = templateSim

    return util.runExperiment(
        individual, experiment, settings, target, experiment[template_name], experiment[template_name].root.timeout, cache
    )


def run(cache):
    "run the parameter estimation"
    searchMethod = cache.settings.get("searchMethod", "NSGA3")
    return cache.search[searchMethod].run(cache, tools, creator)



import CADETMatch.util as util
from CADETMatch.abstract.transform import AbstractTransform


class NormMultTransform(AbstractTransform):
    @property
    def name(self):
        return "norm_mult"

    @property
    def count(self):
        if self.parameter["min"] != self.parameter["max"]:
            return 1
        else:
            return 0

    @property
    def count_extended(self):
        return self.count

    def transform(self):
        if self.count:
            minValue = self.parameter["min"]
            maxValue = self.parameter["max"]

            def trans(i):
                return (i - minValue) / (maxValue - minValue)

            return [
                trans,
            ]
        else:
            return []

    grad_transform = transform

    def untransform_inputorder(self, seq):
        if self.count:
            minValue = self.parameter["min"]
            maxValue = self.parameter["max"]

            values = [
                (maxValue - minValue) * seq[0] + minValue,
            ]
            return values
        else:
            return []

    def untransform(self, seq):
        values = self.untransform_inputorder(seq)
        if self.count:
            headerValues = values
            return values, headerValues
        else:
            return [], []

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        if self.count:
            minValue = self.parameter["min"]
            maxValue = self.parameter["max"]

            values = (maxValue - minValue) * matrix + minValue
            return values
        else:
            return None

    untransform_matrix_inputorder = untransform_matrix

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            locationFrom = self.parameter["locationFrom"]
            locationTo = self.parameter["locationTo"]

            try:
                compFrom = self.parameter["componentFrom"]
                boundFrom = self.parameter["boundFrom"]
                indexFrom = None
            except KeyError:
                indexFrom = self.parameter["indexFrom"]
                boundFrom = None
                compFrom = None
            valueFrom = self.getValue(sim, locationFrom, bound=boundFrom, comp=compFrom, index=indexFrom)

            try:
                compTo = self.parameter["componentTo"]
                boundTo = self.parameter["boundTo"]
                indexTo = None
            except KeyError:
                indexTo = self.parameter["indexTo"]
                boundTo = None
                compTo = None

            if self.count:
                temp = values[0]
            else:
                temp = self.parameter["min"]

            self.setValue(sim, valueFrom * temp, locationTo, bound=boundTo, comp=compTo, index=indexTo)

        if self.count:
            return values, headerValues
        else:
            return [], []

    def getBounds(self):
        if self.count:
            return [0.0,], [1.0,]
        else:
            return None, None

    def getGradBounds(self):
        if self.count:
            return [self.parameter["min"],], [self.parameter["max"],]
        else:
            return None, None

    def getHeaders(self):
        if self.count:
            location = self.parameter["locationTo"]

            try:
                comp = self.parameter["componentTo"]
            except KeyError:
                comp = "None"

            name = location.rsplit("/", 1)[-1]
            bound = self.parameter.get("boundTo", None)
            index = self.parameter.get("indexTo", None)

            headers = []
            if bound is not None:
                headers.append("%s Comp:%s Bound:%s" % (name, comp, bound))
            elif index is not None:
                headers.append("%s Comp:%s Index:%s" % (name, comp, index))
            return headers
        else:
            return []

    def getHeadersActual(self):
        return self.getHeaders()

    def setBounds(self, parameter, lb, ub):
        if self.count:
            parameter["min"] = lb[0]
            parameter["max"] = ub[0]


plugins = {"norm_mult": NormMultTransform}



import CADETMatch.match
import IPython.display as dp
from CADETMatch.cache import Cache
import subprocess
import sys
import pandas
import psutil
import signal


class Match:
    def __init__(self, json_path):
        self.json_path = json_path
        self.cache = Cache()
        self.cache.setup_dir(json_path)
        CADETMatch.match.createDirectories(self.cache, json_path)
        self.cache.setup(json_path)

    def start_sim(self):
        ncpus = psutil.cpu_count(logical=False)
        command = [sys.executable, CADETMatch.match.__file__, str(self.json_path), str(ncpus)]  #'-n', str(ncpus),
        pipe = subprocess.PIPE

        proc = subprocess.Popen(command, stdout=pipe, stderr=subprocess.STDOUT, bufsize=1)

        def signal_handler(sig, frame):
            for child in psutil.Process(proc.pid).children(recursive=True):
                child.kill()
            proc.kill()
            print("Terminating")
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)

        for line in iter(proc.stdout.readline, b""):
            print(line.decode("utf-8"))
        return proc.poll()

    def plot_corner(self):
        process_dir = self.cache.settings["resultsDirProgress"]

        corner = process_dir / "sns_corner.png"
        corner_transform = process_dir / "sns_corner_transform.png"

        if corner.exists():
            print("Corner plot in search space")
            image = dp.Image(filename=corner.as_posix(), embed=True)
            dp.display(image)

        if corner_transform.exists():
            print("Corner plot in original space")
            image = dp.Image(filename=corner_transform.as_posix(), embed=True)
            dp.display(image)

    def plot_best(self):
        meta_dir = self.cache.settings["resultsDirMeta"]
        best, score, best_score = self.get_best()

        for name, col_names in best.items():

            print("Best item %s for meta score(s) %s" % (name, " , ".join(col_names)))
            dp.display(score[name])
            images = meta_dir.glob("%s_*.png" % name)
            for image in images:
                img = dp.Image(filename=image.as_posix(), embed=True)
                dp.display(img)

    def get_best(self):
        "return the best values for each score"
        meta_dir = self.cache.settings["resultsDirMeta"]
        csv = meta_dir / "results.csv"
        data = pandas.read_csv(csv.as_posix())

        best = {}
        score = {}
        best_score = {}

        cols = [("Product Root Score", False), ("Min Score", False), ("Mean Score", False), ("SSE", False)]

        for col_name, order in cols:
            temp = data.sort_values(by=[col_name,], ascending=order)
            head = temp.head(1)
            name = str(head.Name.iloc[0])
            score[name] = head
            best_score[col_name] = head[self.cache.parameter_headers].values[0]
            if name not in best:
                best[name] = [col_name]
            else:
                best[name].append(col_name)

        return best, score, best_score

    def plot_space(self):
        space_dir = self.cache.settings["resultsDirSpace"] / "2d"

        for image in space_dir.glob("*.png"):
            if "1- " in image.name or "SSE" in image.name:
                print(image.name)
                img = dp.Image(filename=image.as_posix(), embed=True)
                dp.display(img)



import importlib
import argparse
import sys
import subprocess
import psutil


def makeParser():
    """Create the CADETMatch module arguments parser."""
    parser = argparse.ArgumentParser(
        description="Starts a parallel version of match using SCOOP.", prog="{0} -m CADETMatch".format(sys.executable),
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--json", "-j", help="Path to JSON file", metavar="JSON")
    parser.add_argument("--match", help="Run CADETMatch with Parameter Estimation and Error Modeling", action="store_true")

    parser.add_argument("--generate_corner", help="Generate corner graphs", action="store_true")

    parser.add_argument("--generate_graphs", help="Generate general graphs excluding 3D", action="store_true")

    parser.add_argument("--generate_graphs_all", help="Generate general graphs including 3D", action="store_true")

    parser.add_argument("--generate_graphs_autocorr", help="Generate autocorrelation graphs", action="store_true")

    parser.add_argument("--generate_spearman", help="Generate spearman graphs", action="store_true")

    parser.add_argument("--generate_mcmc_plot_tube", help="Generate mcmc plot tube graphs", action="store_true")

    parser.add_argument("--generate_mcmc_mixing", help="Generate mcmc mixing graphs", action="store_true")

    parser.add_argument("--generate_mle", help="Generate maximum liklihood", action="store_true")

    parser.add_argument("-n", help="Number of scoop processes to use. Use 1 for debugging", default=0)

    return parser


def run_command(module, json, number_of_jobs, additional=None):
    command = [
        sys.executable,
    ]
    # if int(number_of_jobs) > 0:
    #    command.extend(['-n', str(number_of_jobs)])
    # else:
    #    ncpus = psutil.cpu_count(logical=False)
    #    command.extend(['-n', str(ncpus)])

    command.extend([importlib.util.find_spec(module).origin, str(json)])
    if additional is not None:
        command.extend(additional)

    command.append(str(number_of_jobs))

    rc = subprocess.run(command, bufsize=1)

    return rc.returncode


if __name__ == "__main__":
    parser = makeParser()
    args = parser.parse_args()

    if args.match:
        sys.exit(run_command("CADETMatch.match", args.json, args.n))
    if args.generate_corner:
        sys.exit(run_command("CADETMatch.generate_corner_graphs", args.json, args.n))
    if args.generate_graphs_autocorr:
        sys.exit(run_command("CADETMatch.generate_autocorr_graphs", args.json, args.n))
    if args.generate_mcmc_mixing:
        sys.exit(run_command("CADETMatch.generate_mixing_graphs", args.json, args.n))
    if args.generate_graphs:
        sys.exit(run_command("CADETMatch.generate_graphs", args.json, args.n, ["1"]))
    if args.generate_graphs_all:
        sys.exit(run_command("CADETMatch.generate_graphs", args.json, args.n, ["2"]))
    if args.generate_spearman:
        sys.exit(run_command("CADETMatch.graph_spearman", args.json, args.n, ["end"]))
    if args.generate_mcmc_plot_tube:
        sys.exit(run_command("CADETMatch.mcmc_plot_tube", args.json, args.n))
    if args.generate_mle:
        sys.exit(run_command("CADETMatch.mle", args.json, args.n))



from sklearn.base import clone
import numpy
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import cross_val_score
from sklearn.neighbors import KernelDensity
import SALib.sample.sobol_sequence
import CADETMatch.util as util
from sklearn import preprocessing
import multiprocessing
import itertools
import time

def fit(x):
    bw, kde_params, scores = x
    kde = KernelDensity(**kde_params)
    kde.set_params(bandwidth=bw)
    cross_val = cross_val_score(kde, scores, cv=20)
    return cross_val, bw

def get_bandwidth(kde, scores):
    start = time.time()
    map = util.getMapFunction()
    
    bw = numpy.logspace(-5, 0, 80)
    
    results = list(map(fit, zip(bw, itertools.repeat(kde.get_params()), itertools.repeat(scores))))

    cross_val, bw = zip(*results)

    cross_val = numpy.array(cross_val)
    bw = numpy.array(bw)

    sort_index = numpy.argsort(bw)

    bw = bw[sort_index]
    cross_val = cross_val[sort_index, :]

    scaler = preprocessing.MinMaxScaler()
    shape = cross_val.shape
    cross_val = scaler.fit_transform(cross_val.reshape(-1,1))
    cross_val = cross_val.reshape(shape)

    mean_value = numpy.mean(cross_val, 1)

    bandwidth = interp_bw(bw, mean_value, cross_val)

    multiprocessing.get_logger().info("Bandwidth %s found in %s seconds", bandwidth, time.time() - start)

    store = numpy.array([bw, mean_value]).T

    kde_final = clone(kde)
    kde_final.set_params(bandwidth=bandwidth)

    return kde_final, bandwidth, store

def interp_bw(params, mean_value, values):
    index = numpy.argmax(mean_value)
    params = numpy.array(params)
    mean_value = numpy.array(mean_value)
    
    indexes = numpy.array([index-1, index, index+1])
    x = params[indexes]
    
    X = numpy.tile(x.reshape(-1,1), values.shape[1])
    Y = values[indexes,:]

    X = numpy.squeeze(X.reshape(-1,1))
    Y = numpy.squeeze(Y.reshape(-1,1))
    
    poly, res = numpy.polynomial.Polynomial.fit(X, Y,2, full=True)
    
    bw = poly.deriv().roots()[0]
    
    return bw


import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
import numpy
import numpy.linalg
from addict import Dict
import CADETMatch.smoothing as smoothing
import multiprocessing

name = "DextranShape"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True
    settings.count = 2
    return settings


def run(sim_data, feature):
    "special score designed for dextran. This looks at only the front side of the peak up to the maximum slope and pins a value at the elbow in addition to the top"
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)

    exp_time_zero = feature["exp_time_zero"]
    exp_data_zero = feature["exp_data_zero"]

    sim_spline, sim_data_zero_sse = cut_front(
        sim_time_values,
        sim_data_values,
        feature["min_value_front"],
        feature["max_value_front"],
        feature["critical_frequency"],
        feature["smoothing_factor"],
    )

    if sim_spline is None:
        return None

    pearson, diff_time = score.pearson_spline_fun(exp_time_zero, exp_data_zero, sim_spline)

    exp_data_zero_sse = feature["exp_data_zero_sse"]

    temp = [
        pearson,
        feature["offsetTimeFunction"](numpy.abs(diff_time)),
    ]

    data = (
        temp,
        util.sse(sim_data_zero_sse, exp_data_zero_sse),
        len(sim_data_zero_sse),
        sim_time_values,
        sim_data_zero_sse,
        exp_data_zero_sse,
        [1.0 - i for i in temp],
    )
    return data


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    # change the stop point to be where the max positive slope is along the searched interval
    name = "%s_%s" % (sim.root.experiment_name, feature["name"])
    exp_time_zero, exp_data_zero, exp_data_zero_sse, min_time, min_value, max_time, max_value, s, crit_fs, crit_fs_der = cut_front_find(
        selectedTimes, selectedValues, name, cache
    )

    multiprocessing.get_logger().info("Dextran %s  start: %s   stop: %s  max value: %s", name, min_time, max_time, max_value)

    temp["min_time"] = feature["start"]
    temp["max_time"] = feature["stop"]

    temp["min_time_front"] = min_time
    temp["min_value_front"] = min_value
    temp["max_time_front"] = max_time
    temp["max_value_front"] = max_value

    temp["exp_time_zero"] = exp_time_zero
    temp["exp_data_zero"] = exp_data_zero
    temp["exp_data_zero_sse"] = exp_data_zero_sse

    decay = feature.get("decay", 1)

    if decay:
        temp["offsetTimeFunction"] = score.time_function_decay(feature["time"][-1])
    else:
        temp["offsetTimeFunction"] = score.time_function(feature["time"][-1], 0.1 * CV_time)

    temp["peak_max"] = max_value
    temp["smoothing_factor"] = s
    temp["critical_frequency"] = crit_fs
    temp["critical_frequency_der"] = crit_fs_der
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = [
        "%s_Shape" % name,
        "%s_Time" % name,
    ]
    return temp


def cut_front_find(times, values, name, cache):
    s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(times, values, name, cache)
    smooth_value, values_der = smoothing.full_smooth(times, values, crit_fs, s, crit_fs_der)

    spline_der = scipy.interpolate.InterpolatedUnivariateSpline(times, values_der, ext=1)
    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, smooth_value, ext=1)

    min_time, min_value, max_time, max_value = score.find_cuts(times, smooth_value, spline, spline_der)

    # resample to 100 points/second
    needed_points = int((times[-1] - times[0]) * 100)

    new_times = numpy.linspace(times[0], times[-1], needed_points)
    new_values = spline(new_times)

    exp_data_zero_sse, exp_min_time, exp_min_value, exp_max_time, exp_max_value = score.cut_zero(times, smooth_value, min_value, max_value)

    if exp_min_time is not None:
        min_time = exp_min_time
        min_value = exp_min_value
        max_time = exp_max_time
        max_value = exp_max_value

    exp_data_zero, _, _, _, _ = score.cut_zero(new_times, new_values, min_value, max_value)

    return (new_times, exp_data_zero, exp_data_zero_sse, min_time, min_value, max_time, max_value, s, crit_fs, crit_fs_der)


def cut_front(times, values, min_value, max_value, crit_fs, s):
    max_index = numpy.argmax(values >= max_value)

    if numpy.max(values) < min_value:
        return None, None

    if max_index == 0:
        # no point high enough was found so use the highest point
        s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(times, values, None, None)
        max_index = numpy.argmax(values)
        max_value = values[max_index]

        if min_value >= max_value:
            min_value = 1e-3 * max_value

    smooth_value = smoothing.smooth_data(times, values, crit_fs, s)

    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, smooth_value, ext=1)

    max_time = score.find_target(spline, max_value, times, smooth_value)

    min_time = score.find_target(spline, min_value, times, smooth_value)

    if max_time is None or min_time is None:
        # result is truly garbage, this means the shape is so distorted compared to the real system that even the spline is not accurate
        # this almost always happens when the peak is so sharp it only spans a few time points
        # it also means the rest of the optimization is not needed
        return None, None

    needed_points = int((max_time - min_time) * 10)
    if needed_points > 10:
        new_times_spline = numpy.linspace(min_time, max_time, needed_points)
        new_values_spline = spline(new_times_spline)

        spline = scipy.interpolate.InterpolatedUnivariateSpline(new_times_spline, new_values_spline, ext=1)

        return spline, score.cut_zero(times, smooth_value, min_value, max_value)[0]
    else:
        return None,None



from deap import base
import numpy


class Fitness2(base.Fitness):
    def dominates(self, other, obj=slice(None)):
        """Return true if each objective of *self* is not strictly worse than
        the corresponding objective of *other* and at least one objective is
        strictly better.
        :param obj: Slice indicating on which objectives the domination is
                    tested. The default value is `slice(None)`, representing
                    every objectives.
        """
        not_equal = False

        self_weight = numpy.array(self.wvalues[obj])
        other_weight = numpy.array(other.wvalues[obj])

        bools = self_weight >= other_weight
        greater_bools = self_weight > other_weight

        if numpy.all(bools) and numpy.sum(greater_bools) >= 2:
            not_equal = True
        else:
            not_equal = False
        return not_equal


class Fitness3(base.Fitness):
    def dominates(self, other, obj=slice(None)):
        """Return true if each objective of *self* is not strictly worse than
        the corresponding objective of *other* and at least one objective is
        strictly better.
        :param obj: Slice indicating on which objectives the domination is
                    tested. The default value is `slice(None)`, representing
                    every objectives.
        """
        not_equal = False

        self_weight = numpy.array(self.wvalues[obj])
        other_weight = numpy.array(other.wvalues[obj])

        bools = self_weight >= other_weight
        greater_bools = self_weight > other_weight

        if numpy.all(bools) and numpy.sum(greater_bools) >= 3:
            not_equal = True
        else:
            not_equal = False
        return not_equal



import CADETMatch.util as util
from CADETMatch.abstract.transform import AbstractTransform
import numpy


class NullTransform(AbstractTransform):
    @property
    def name(self):
        return "null"

    @property
    def count(self):
        return 1

    @property
    def count_extended(self):
        return self.count

    def transform(self):
        def trans(i):
            return i

        return [
            trans,
        ]

    grad_transform = transform


    def untransform_inputorder(self, seq):
        values = [
            seq[0],
        ]
        return values

    def untransform(self, seq):
        values = self.untransform_inputorder(seq)
        headerValues = values
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = numpy.zeros(matrix.shape)
        values[:, 0] = matrix[:, 0]
        return values

    untransform_matrix_inputorder = untransform_matrix

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            location = self.parameter["location"]

            try:
                comp = self.parameter["component"]
                bound = self.parameter["bound"]
                index = None
            except KeyError:
                index = self.parameter["index"]
                bound = None

            if bound is not None:
                unit = self.getUnit(location)
                boundOffset = util.getBoundOffset(sim.root.input.model[unit])

                if comp == -1:
                    position = ()
                    sim[location.lower()] = values[0]
                else:
                    position = boundOffset[comp] + bound
                    sim[location.lower()][position] = values[0]

            if index is not None:
                sim[location.lower()][index] = values[0]
            return values, headerValues

    def getBounds(self):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        return [minValue,], [maxValue,]

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]

    def getHeaders(self):
        location = self.parameter["location"]

        try:
            comp = self.parameter["component"]
        except KeyError:
            comp = "None"

        name = location.rsplit("/", 1)[-1]
        bound = self.parameter.get("bound", None)
        index = self.parameter.get("index", None)

        headers = []
        if bound is not None:
            headers.append("%s Comp:%s Bound:%s" % (name, comp, bound))
        elif index is not None:
            headers.append("%s Comp:%s Index:%s" % (name, comp, index))
        return headers

    def getHeadersActual(self):
        return self.getHeaders()

    def setBounds(self, parameter, lb, ub):
        parameter["min"] = lb[0]
        parameter["max"] = ub[0]


class NormTransform(NullTransform):
    @property
    def name(self):
        return "norm"

    def transform(self):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        def trans(i):
            return (i - minValue) / (maxValue - minValue)

        return [
            trans,
        ]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        values = [
            (maxValue - minValue) * seq[0] + minValue,
        ]
        return values


    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        values = (maxValue - minValue) * matrix + minValue
        return values

    untransform_matrix_inputorder = untransform_matrix

    def getBounds(self):
        return [0.0,], [1.0,]

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]


plugins = {"norm": NormTransform, "null": NullTransform}



import sys

import matplotlib
import matplotlib.style as mplstyle
mplstyle.use('fast')

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D

from CADETMatch.cache import cache

from pathlib import Path
import pandas
import numpy
import scipy.interpolate
import itertools

from cadet import Cadet, H5
from addict import Dict

# parallelization
import multiprocessing

import os
import warnings
import CADETMatch.util as util
import logging
import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.smoothing as smoothing

saltIsotherms = {
    b"STERIC_MASS_ACTION",
    b"SELF_ASSOCIATION",
    b"MULTISTATE_STERIC_MASS_ACTION",
    b"SIMPLE_MULTISTATE_STERIC_MASS_ACTION",
    b"BI_STERIC_MASS_ACTION",
}


from matplotlib.colors import ListedColormap
import matplotlib.cm

cmap = matplotlib.cm.winter

# Get the colormap colors
my_cmap = cmap(numpy.arange(cmap.N))

# Set alpha
my_cmap[:, -1] = 1.0

# Create new colormap
my_cmap = ListedColormap(my_cmap)

cm_plot = matplotlib.cm.gist_rainbow


def get_color(idx, max_colors, cmap):
    return cmap(1.0 * float(idx) / max_colors)


def main(map_function):
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "graph.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("graphing directory %s", os.getcwd())

    fullGeneration = int(sys.argv[2])

    # full generation 1 = 2D, 2 = 2D + 3D

    graphMeta(cache, map_function)
    graphProgress(cache, map_function)

    if fullGeneration:
        graphSpace(fullGeneration, cache, map_function)
        graphExperiments(cache, map_function)


def graphDistance(cache, map_function):

    resultDir = Path(cache.settings["resultsDir"])
    result_h5 = resultDir / "result.h5"

    output_distance = cache.settings["resultsDirSpace"] / "distance"
    output_distance_meta = output_distance / "meta"

    output_distance.mkdir(parents=True, exist_ok=True)

    output_distance_meta.mkdir(parents=True, exist_ok=True)

    parameter_headers_actual = cache.parameter_headers_actual
    score_headers = cache.score_headers
    meta_headers = cache.meta_headers

    if result_h5.exists():
        data = H5()
        data.filename = result_h5.as_posix()
        data.load()

        temp = []

        if "mean" in data.root and "confidence" in data.root and "distance_correct" in data.root:
            for idx_parameter, parameter in enumerate(parameter_headers_actual):
                for idx_score, score in enumerate(score_headers):
                    temp.append(
                        (
                            output_distance,
                            parameter,
                            idx_parameter,
                            score,
                            idx_score,
                            data.root.distance_correct[:, idx_parameter],
                            data.root.output[:, idx_score],
                            data.root.mean[:, idx_parameter],
                            data.root.confidence[:, idx_parameter],
                        )
                    )

            for idx_parameter, parameter in enumerate(parameter_headers_actual):
                for idx_score, score in enumerate(meta_headers):
                    temp.append(
                        (
                            output_distance_meta,
                            parameter,
                            idx_parameter,
                            score,
                            idx_score,
                            data.root.distance_correct[:, idx_parameter],
                            data["output_meta"][:, idx_score],
                            data.root.mean[:, idx_parameter],
                            data.root.confidence[:, idx_parameter],
                        )
                    )

            list(map_function(plot_2d_scatter, temp))


def plot_2d_scatter(args):
    output_directory, parameter, parameter_idx, score, score_idx, parameter_data, score_data, mean_data, confidence_data = args
    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(parameter_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))

    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(mean_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s_mean.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))

    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(confidence_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s_confidence.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))


def graphExperiments(cache, map_function):
    multiprocessing.get_logger().info("starting simulation graphs")
    directory = Path(cache.settings["resultsDirEvo"])

    # find all items in directory
    pathsH5 = directory.glob("*.h5")
    pathsPNG = directory.glob("*.png")

    # make set of items based on removing everything after _
    existsH5 = {str(path.name).split("_", 1)[0] for path in pathsH5}
    existsPNG = {str(path.name).split("_", 1)[0] for path in pathsPNG}

    toGenerate = existsH5 - existsPNG

    args = zip(
        toGenerate, itertools.repeat(sys.argv[1]), itertools.repeat(cache.settings["resultsDirEvo"]), itertools.repeat("%s_%s_EVO.png")
    )

    list(map_function(plotExperiments, args))
    multiprocessing.get_logger().info("ending simulation graphs")


def graphMeta(cache, map_function):
    multiprocessing.get_logger().info("starting meta graphs")
    directory = Path(cache.settings["resultsDirMeta"])

    # find all items in directory
    pathsH5 = directory.glob("*.h5")
    pathsPNG = directory.glob("*.png")

    # make set of items based on removing everything after _
    existsH5 = {str(path.name).split("_", 1)[0] for path in pathsH5}
    existsPNG = {str(path.name).split("_", 1)[0] for path in pathsPNG}

    toGenerate = existsH5 - existsPNG

    args = zip(
        toGenerate, itertools.repeat(sys.argv[1]), itertools.repeat(cache.settings["resultsDirMeta"]), itertools.repeat("%s_%s_meta.png")
    )

    list(map_function(plotExperiments, args))
    multiprocessing.get_logger().info("ending meta graphs")


def graph_simulation_unit(simulation, unit, graph):
    ncomp = int(simulation.root.input.model[unit].ncomp)
    isotherm = bytes(simulation.root.input.model[unit].adsorption_model)

    # This does not work correctly since we could be reading an outlet or tubing that does not have an isotherm
    # hasSalt = isotherm in saltIsotherms

    solution_times = simulation.root.output.solution.solution_times

    comps = []

    hasColumn = any("column" in i for i in simulation.root.output.solution[unit].keys())
    hasPort = any("port" in i for i in simulation.root.output.solution[unit].keys())

    if hasColumn:
        for i in range(ncomp):
            comps.append(simulation.root.output.solution[unit]["solution_column_outlet_comp_%03d" % i])
    elif hasPort:
        for i in range(ncomp):
            comps.append(simulation.root.output.solution[unit]["solution_outlet_port_000_comp_%03d" % i])
    else:
        for i in range(ncomp):
            comps.append(simulation.root.output.solution[unit]["solution_outlet_comp_%03d" % i])

    # for now lets assume that if your first component is much higher concentration than the other components then the first component is salt
    max_comps = [max(comp) for comp in comps]

    hasSalt = any([max_comps[0] > (comp * 10) for comp in max_comps[1:]])

    if hasSalt:
        graph.set_title("Output %s" % unit)
        graph.plot(solution_times, comps[0], "b-", label="Salt")
        graph.set_xlabel("time (s)")

        # Make the y-axis label, ticks and tick labels match the line color.
        graph.set_ylabel("mM Salt", color="b")
        graph.tick_params("y", colors="b")

        axis2 = graph.twinx()
        for idx, comp in enumerate(comps[1:]):
            axis2.plot(solution_times, comp, "-", color=get_color(idx, len(comps) - 1, cm_plot), label="P%s" % idx)
        axis2.set_ylabel("mM Protein", color="r")
        axis2.tick_params("y", colors="r")

        lines, labels = graph.get_legend_handles_labels()
        lines2, labels2 = axis2.get_legend_handles_labels()
        axis2.legend(lines + lines2, labels + labels2, loc=0)
    else:
        graph.set_title("Output %s" % unit)

        # colors = ['r', 'g', 'c', 'm', 'y', 'k']
        for idx, comp in enumerate(comps):
            graph.plot(solution_times, comp, "-", color=get_color(idx, len(comps), cm_plot), label="P%s" % idx)
        graph.set_ylabel("mM Protein", color="r")
        graph.tick_params("y", colors="r")
        graph.set_xlabel("time (s)")

        lines, labels = graph.get_legend_handles_labels()
        graph.legend(lines, labels, loc=0)


def convert_frac(start, stop, values):
    x = []
    y = []
    for xl, xu, val in zip(start, stop, values):
        x.append(xl)
        y.append(val)
        x.append(xu)
        y.append(val)
    return x, y


def plotExperiments(args):
    save_name_base, json_path, directory, file_pattern = args
    if json_path != cache.json_path:
        cache.setup(json_path)

    target = cache.target
    settings = cache.settings
    for experiment in settings["experiments"]:
        experimentName = experiment["name"]

        dst = Path(directory, file_pattern % (save_name_base, experimentName))

        if dst.exists():
            # this item has already been plotted, this means that plots are occurring faster than generations are so kill this version
            sys.exit()

        units_used = cache.target[experimentName]["units_used"]

        numPlotsSeq = [len(units_used)]
        # Shape and ShapeDecay have a chromatogram + derivative
        for feature in experiment["features"]:
            if feature["type"] in ("Shape", "ShapeDecay", "ShapeFront", "ShapeBack"):
                numPlotsSeq.append(2)
            elif feature["type"] in ("AbsoluteTime", "AbsoluteHeight"):
                pass
            else:
                numPlotsSeq.append(1)

        numPlots = sum(numPlotsSeq)

        exp_time = target[experimentName]["time"]
        exp_value = target[experimentName]["value"]

        fig = figure.Figure(figsize=[24, 12 * numPlots])
        canvas = FigureCanvas(fig)

        simulation = Cadet()
        h5_path = Path(directory) / (file_pattern.replace("png", "h5") % (save_name_base, experimentName))
        simulation.filename = h5_path.as_posix()
        simulation.load()

        results = {}
        results["simulation"] = simulation

        for idx, unit in enumerate(units_used):
            graph_simulation_unit(simulation, unit, fig.add_subplot(numPlots, 1, idx + 1))

        graphIdx = idx + 2
        for idx, feature in enumerate(experiment["features"]):
            featureName = feature["name"]
            featureType = feature["type"]

            feat = target[experimentName][featureName]

            selected = feat["selected"]
            exp_time = feat["time"][selected]
            exp_value = feat["value"][selected]

            sim_time, sim_value = get_times_values(simulation, target[experimentName][featureName])

            if featureType in (
                "similarity",
                "similarityDecay",
                "similarityHybrid",
                "similarityHybrid2",
                "similarityHybrid2_spline",
                "similarityHybridDecay",
                "similarityHybridDecay2",
                "curve",
                "breakthrough",
                "dextran",
                "dextranHybrid",
                "dextranHybrid2",
                "dextranHybrid2_spline",
                "similarityCross",
                "similarityCrossDecay",
                "breakthroughCross",
                "SSE",
                "LogSSE",
                "breakthroughHybrid",
                "breakthroughHybrid2",
                "Shape",
                "ShapeDecay",
                "Dextran",
                "DextranAngle",
                "DextranTest",
                "DextranQuad",
                "Dextran3",
                "DextranShape",
                "ShapeDecaySimple",
                "ShapeSimple",
                "DextranSSE",
                "ShapeFront",
                "ShapeBack",
                "ShapeNoDer",
                "ShapeDecayNoDer",
            ):

                graph = fig.add_subplot(numPlots, 1, graphIdx)  # additional +1 added due to the overview plot
                graph.plot(sim_time, sim_value, "r--", label="Simulation")
                graph.plot(exp_time, exp_value, "g:", label="Experiment")
                graphIdx += 1

            if featureType in (
                "derivative_similarity",
                "derivative_similarity_hybrid",
                "derivative_similarity_hybrid2",
                "derivative_similarity_cross",
                "derivative_similarity_cross_alt",
                "derivative_similarity_hybrid2_spline",
                "similarityHybridDecay2_spline",
                "Shape",
                "ShapeDecay",
                "ShapeFront",
                "ShapeBack",
            ):
                exp_spline = smoothing.smooth_data_derivative(
                    exp_time, exp_value, feat["critical_frequency"], feat["smoothing_factor"], feat["critical_frequency_der"]
                )
                sim_spline = smoothing.smooth_data_derivative(
                    sim_time, sim_value, feat["critical_frequency"], feat["smoothing_factor"], feat["critical_frequency_der"]
                )

                graph = fig.add_subplot(numPlots, 1, graphIdx)  # additional +1 added due to the overview plot
                graph.plot(sim_time, sim_spline, "r--", label="Simulation")
                graph.plot(exp_time, exp_spline, "g:", label="Experiment")
                graphIdx += 1

            if featureType in (
                "fractionation",
                "fractionationCombine",
                "fractionationMeanVariance",
                "fractionationMoment",
                "fractionationSlide",
                "fractionationSSE",
            ):
                cache.scores[featureType].run(results, feat)

                graph_exp = results["graph_exp"]
                graph_sim = results["graph_sim"]
                graph_sim_offset = results.get("graph_sim_offset", {})

                findMax = 0
                max_comp = {}
                for idx, (key, value) in enumerate(graph_sim.items()):
                    (time_start, time_stop, values) = zip(*value)
                    max_value = max(values)
                    findMax = max(findMax, max_value)
                    max_comp[key] = max_value

                for idx, (key, value) in enumerate(graph_exp.items()):
                    (time_start, time_stop, values) = zip(*value)
                    max_value = max(values)
                    findMax = max(findMax, max_value)
                    max_comp[key] = max(max_comp[key], max_value)

                graph = fig.add_subplot(numPlots, 1, graphIdx)  # additional +1 added due to the overview plot
                factors = []
                for idx, (key, value) in enumerate(graph_sim.items()):
                    (time_start, time_stop, values) = zip(*value)
                    values = numpy.array(values)
                    mult = 1.0
                    if max(values) < (0.2 * findMax):
                        mult = findMax / (2 * max_comp[key])
                        values = values * mult
                    factors.append(mult)

                    time_offset = graph_sim_offset.get(key, None)
                    if time_offset is not None:
                        if time_offset > 0:
                            time_str = "time early (s): %.3g" % abs(time_offset)
                        else:
                            time_str = "time late (s): %.3g" % abs(time_offset)
                    else:
                        time_str = ""

                    label = "Simulation Comp: %s Mult:%.2f %s" % (key, mult, time_str)

                    x, y = convert_frac(time_start, time_stop, values)

                    graph.plot(x, y, "--", color=get_color(idx, len(graph_sim), cm_plot), label=label)

                for idx, (key, value) in enumerate(graph_exp.items()):
                    (time_start, time_stop, values) = zip(*value)
                    values = numpy.array(values)
                    mult = factors[idx]
                    values = values * mult

                    x, y = convert_frac(time_start, time_stop, values)

                    graph.plot(
                        x, y, ":", color=get_color(idx, len(graph_sim), cm_plot), label="Experiment Comp: %s Mult:%.2f" % (key, mult)
                    )
                graphIdx += 1
            graph.legend()
        fig.set_size_inches((24, 12 * numPlots))
        fig.savefig(str(dst))


def graphSpace(fullGeneration, cache, map_function):
    multiprocessing.get_logger().info("starting space graphs")
    progress_path = Path(cache.settings["resultsDirBase"]) / "result.h5"

    output_2d = cache.settings["resultsDirSpace"] / "2d"
    output_3d = cache.settings["resultsDirSpace"] / "3d"
    output_prob = cache.settings["resultsDirSpace"] / "probability"

    output_2d.mkdir(parents=True, exist_ok=True)
    output_3d.mkdir(parents=True, exist_ok=True)
    output_prob.mkdir(parents=True, exist_ok=True)

    results = H5()
    results.filename = progress_path.as_posix()

    results.load(paths=["/output", "/output_meta", "/is_extended_input", "/probability"])

    if results.root.is_extended_input:
        results.load(paths=["/input_transform_extended"], update=True)
        input = results.root.input_transform_extended
    else:
        results.load(paths=["/input_transform"], update=True)
        input = results.root.input_transform

    if isinstance(results.root.probability, Dict):
        probability = None
    else:
        probability = numpy.squeeze(results.root.probability)
        probability_idx = numpy.max(probability) / probability < 1e5
        probability = probability[probability_idx]

    output = results.root.output
    output_meta = results.root.output_meta

    input_headers = cache.parameter_headers
    output_headers = cache.score_headers
    meta_headers = cache.meta_headers

    input_indexes = list(range(len(input_headers)))
    output_indexes = list(range(len(output_headers)))
    meta_indexes = list(range(len(meta_headers)))

    comp_two = list(itertools.combinations(input_indexes, 2))
    comp_one = list(itertools.combinations(input_indexes, 1))

    # 2d plots
    if fullGeneration >= 1:
        multiprocessing.get_logger().info("starting 2d space graphs")
        graphDistance(cache, map_function)

        seq = []
        for (x,), y in itertools.product(comp_one, output_indexes):
            seq.append([output_2d.as_posix(), input_headers[x], output_headers[y], input[:, x], output[:, y]])

        for (x,), y in itertools.product(comp_one, meta_indexes):
            seq.append([output_2d.as_posix(), input_headers[x], meta_headers[y], input[:, x], output_meta[:, y]])

        if probability is not None:
            for (x,) in comp_one:
                seq.append([output_prob.as_posix(), input_headers[x], "probability", input[probability_idx, x], probability])

        list(map_function(plot_2d, seq))
        multiprocessing.get_logger().info("ending 2d space graphs")

    # 3d plots
    if fullGeneration >= 2:
        multiprocessing.get_logger().info("starting 3d space graphs")
        seq = []
        for (x, y), z in itertools.product(comp_two, output_indexes):
            seq.append(
                [output_3d.as_posix(), input_headers[x], input_headers[y], output_headers[z], input[:, x], input[:, y], output[:, z]]
            )

        for (x, y), z in itertools.product(comp_two, meta_indexes):
            seq.append(
                [output_3d.as_posix(), input_headers[x], input_headers[y], meta_headers[z], input[:, x], input[:, y], output_meta[:, z]]
            )
        list(map_function(plot_3d, seq))
        multiprocessing.get_logger().info("ending 3d space graphs")
    multiprocessing.get_logger().info("ending space graphs")


def plot_3d(arg):
    "This leaks memory and is run in a separate short-lived process, do not integrate into the matching main process"
    directory_path, header1, header2, header3, data1, data2, scores = arg
    directory = Path(directory_path)

    x_str = "%s"
    x = data1
    if max(x) / min(x) > 100:
        x_str = "log10(%s)"
        x = numpy.log10(x)

    y_str = "%s"
    y = data2
    if max(y) / min(y) > 100:
        y_str = "log10(%s)"
        y = numpy.log10(y)

    z_str = "%s"
    z = scores
    if max(z) / min(z) > 100:
        z_str = "log10(%s)"
        z = numpy.log10(z)

    fig = figure.Figure(figsize=[20, 20])
    canvas = FigureCanvas(fig)
    ax = fig.add_subplot(111, projection="3d")
    ax.scatter(x, y, z, c=z, cmap=my_cmap)
    ax.set_xlabel(x_str % header1)
    ax.set_ylabel(y_str % header2)
    ax.set_zlabel(z_str % header3)
    filename = "%s_%s_%s.png" % (header1, header2, header3)
    filename = filename.replace(":", "_").replace("/", "_")
    fig.savefig(str(directory / filename))


def plot_2d(arg):
    directory_path, header_x, scoreName, data, scores = arg

    multiprocessing.get_logger().info("2d graph %s %s start" % (header_x, scoreName))

    plot_2d_single(directory_path, header_x, scoreName, data, scores)

    multiprocessing.get_logger().info("2d graph %s %s stop" % (header_x, scoreName))


def plot_2d_single(directory_path, header_x, scoreName, data, scores):
    directory = Path(directory_path)

    fig = figure.Figure(figsize=[10, 10])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    format = "%s"
    if numpy.max(data) / numpy.min(data[data > 0]) > 100.0:
        keep = data > 0
        data = numpy.log10(data[keep])
        scores = scores[keep]
        format = "log10(%s)"

    format_scores = "%s"
    if numpy.max(scores) / numpy.min(scores[scores > 0]) > 100.0:
        keep = scores > 0
        scores = numpy.log10(scores[keep])
        data = data[keep]
        format_scores = "log10(%s)"

    data_norm = (data - numpy.min(data)) / (numpy.max(data) - numpy.min(data))
    scores_norm = (scores - numpy.min(scores)) / (numpy.max(scores) - numpy.min(scores))
    scatter_data = numpy.array([data_norm, scores_norm]).T
    scatter_data_round = numpy.round(scatter_data, 2)
    scatter_data_round, idx, counts = numpy.unique(scatter_data_round, axis=0, return_index=True, return_counts=True)

    counts_norm = (counts - numpy.min(counts)) / (numpy.max(counts) - numpy.min(counts)) * 100 + 10

    graph.scatter(data[idx], scores[idx], c=scores[idx], s=counts_norm, cmap=cmap)

    graph.set_xlabel(format % header_x)
    graph.set_ylabel(format_scores % scoreName)
    graph.set_xlim(min(data), max(data), auto=True)
    filename = "%s_%s.png" % (header_x, scoreName)
    filename = filename.replace(":", "_").replace("/", "_")

    fig.savefig(str(directory / filename))


def graphProgress(cache, map_function):
    multiprocessing.get_logger().info("starting progress graphs")
    results = Path(cache.settings["resultsDirBase"])
    progress = results / "progress.csv"

    df = pandas.read_csv(progress)

    output = cache.settings["resultsDirProgress"]

    x = [
        "Generation",
    ]
    y = ["Meta Front", "Meta Min", "Meta Product", "Meta Mean", "Meta SSE", "Meta RMSE"]

    temp = []
    for x, y in itertools.product(x, y):
        if x != y:
            temp.append((df[[x, y]], output))

    list(map_function(singleGraphProgress, temp))
    multiprocessing.get_logger().info("ending progress graphs")


def singleGraphProgress(arg):
    df, output = arg

    i, j = list(df)

    fig = figure.Figure(figsize=(10, 10))
    canvas = FigureCanvas(fig)

    graph = fig.add_subplot(1, 1, 1)

    x = df[i].to_numpy()
    y = df[j].to_numpy()

    title_str = "%s vs %s"
    y_label = "%s"
    if numpy.max(y) / numpy.min(y) > 100.0:
        y = numpy.log10(y)
        title_str = "%s vs log10(%s)"
        y_label = "log10(%s)"

    graph.plot(x, y)
    y_max = max(y)
    y_min = min(y)
    graph.set_ylim((1.1 * y_min, 1.1 * y_max))
    graph.set_title(title_str % (i, j))
    graph.set_xlabel(i)
    graph.set_ylabel(y_label % j)

    filename = "%s vs %s.png" % (i, j)
    file_path = output / filename
    fig.savefig(str(file_path))


def get_times_values(simulation, target, selected=None):

    times = simulation.root.output.solution.solution_times

    isotherm = target["isotherm"]

    if isinstance(isotherm, list):
        values = numpy.sum([simulation[i] for i in isotherm], 0)
    else:
        values = simulation[isotherm]

    if selected is None:
        selected = target["selected"]

    return times[selected], values[selected] * target["factor"]


if __name__ == "__main__":
    map_function = util.getMapFunction()
    main(map_function)
    sys.exit()



import CADETMatch.util as util
import numpy
import CADETMatch.calc_coeff as calc_coeff
from CADETMatch.abstract.transform import AbstractTransform


class LinearTransform(AbstractTransform):
    @property
    def name(self):
        return "linear"

    @property
    def count(self):
        return 2

    @property
    def count_extended(self):
        return self.count

    def transform(self):
        def trans_a(i):
            return i

        def trans_b(i):
            return i

        return [trans_a, trans_b]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        values = [seq[0], seq[1]]
        return values

    def untransform(self, seq):
        values = self.untransform_inputorder(seq)
        headerValues = values
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = numpy.array(matrix)
        return values

    untransform_matrix_inputorder = untransform_matrix

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            location = self.parameter["location"]

            minX = self.parameter["minX"]
            maxX = self.parameter["maxX"]

            x_name = self.parameter["x_name"]

            x_value = experiment[x_name]

            slope, intercept = calc_coeff.linear_coeff(minX, values[0], maxX, values[1])

            value = calc_coeff.linear(x_value, slope, intercept)

            try:
                comp = self.parameter["component"]
                bound = self.parameter["bound"]
                index = None
            except KeyError:
                index = self.parameter["index"]
                bound = None

            if bound is not None:
                unit = self.getUnit(location)
                boundOffset = util.getBoundOffset(sim.root.input.model[unit])

                if comp == -1:
                    position = ()
                    sim[location.lower()] = value
                else:
                    position = boundOffset[comp] + bound
                    sim[location.lower()][position] = value

            if index is not None:
                sim[location.lower()][index] = value

        return values, headerValues

    def getBounds(self):
        minLower = self.parameter["minLower"]
        maxLower = self.parameter["maxLower"]
        minUpper = self.parameter["minUpper"]
        maxUpper = self.parameter["maxUpper"]

        minValues = [minLower, minUpper]
        maxValues = [maxLower, maxUpper]

        return minValues, maxValues

    def getGradBounds(self):
        return self.getBounds()

    def getHeaders(self):
        bound = self.parameter["bound"]
        comp = self.parameter["component"]

        headers = []
        headers.append("Lower Comp:%s Bound:%s" % (comp, bound))
        headers.append("Upper Comp:%s Bound:%s" % (comp, bound))
        return headers

    def getHeadersActual(self):
        return self.getHeaders()

    def setBounds(self, parameter, lb, ub):
        parameter["minLower"] = lb[0]
        parameter["maxLower"] = ub[0]
        parameter["minUpper"] = lb[1]
        parameter["maxUpper"] = ub[1]


class NormLinearTransform(LinearTransform):
    @property
    def name(self):
        return "norm_linear"

    def transform(self):
        minLower = self.parameter["minLower"]
        maxLower = self.parameter["maxLower"]
        minUpper = self.parameter["minUpper"]
        maxUpper = self.parameter["maxUpper"]

        def trans_a(i):
            return (i - minLower) / (maxLower - minLower)

        def trans_b(i):
            return (i - minUpper) / (maxUpper - minUpper)

        return [trans_a, trans_b]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        minLower = self.parameter["minLower"]
        maxLower = self.parameter["maxLower"]
        minUpper = self.parameter["minUpper"]
        maxUpper = self.parameter["maxUpper"]
        minX = self.parameter["minX"]
        maxX = self.parameter["maxX"]

        minValues = numpy.array([minLower, minUpper])
        maxValues = numpy.array([maxLower, maxUpper])

        values = numpy.array(seq)

        values = (maxValues - minValues) * values + minValues
        return values

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        minLower = self.parameter["minLower"]
        maxLower = self.parameter["maxLower"]
        minUpper = self.parameter["minUpper"]
        maxUpper = self.parameter["maxUpper"]

        minValues = numpy.array([minLower, minUpper])
        maxValues = numpy.array([maxLower, maxUpper])

        values = (maxValues - minValues) * values + minValues

        return values

    untransform_matrix_inputorder = untransform_matrix

    def getBounds(self):
        return [0.0, 0.0], [1.0, 1.0]

    def getGradBounds(self):
        minLower = self.parameter["minLower"]
        maxLower = self.parameter["maxLower"]
        minUpper = self.parameter["minUpper"]
        maxUpper = self.parameter["maxUpper"]
        return [minLower, minUpper], [maxLower, maxUpper]


plugins = {"norm_linear": NormLinearTransform, "linear": LinearTransform}



import CADETMatch.util as util
import CADETMatch.calc_coeff as calc_coeff
import CADETMatch.score as score
import numpy
from addict import Dict
import scipy

name = "AbsoluteTime"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = False
    settings.count = 1
    return settings


"""This score is NOT for optimization. It is needed for the MCMC algorithm in order to handle assymetric distributions"""


def run(sim_data, feature):
    "similarity, value, start stop"
    selected = feature["selected"]
    exp_data_values = feature["value"][selected]
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)

    ys = [0.0, 1.0]
    xs = [0.0, feature["exp_time"]]

    a, b = calc_coeff.linear_coeff(xs[0], ys[0], xs[1], ys[1])

    sim_time = find_time(sim_time_values, sim_data_values)

    value = calc_coeff.linear(sim_time, a, b)

    temp = [
        value,
    ]

    return (
        temp,
        util.sse(sim_data_values, exp_data_values),
        len(sim_data_values),
        sim_time_values,
        sim_data_values,
        exp_data_values,
        [1.0 - i for i in temp],
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    exp_time = find_time(selectedTimes, selectedValues)

    temp = {}
    temp["exp_time"] = exp_time
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = ["%s_Time" % name]
    return temp


def find_time(times, values):
    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, values, ext=3)

    arg_max = numpy.argmax(values)
    max_value = values[arg_max]
    search_value = 0.9 * max_value

    time = score.find_target(spline, search_value, times[:arg_max], values[:arg_max])

    return time



import CADETMatch.util as util
import CADETMatch.calc_coeff as calc_coeff
import CADETMatch.score as score
from addict import Dict

name = "Ceiling"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True
    settings.count = 1
    return settings


def run(sim_data, feature):
    "similarity, value, start stop"
    selected = feature["selected"]
    exp_data_values = feature["value"][selected]
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)

    max_value = max(sim_data_values)

    if max_value <= feature["max_value"]:
        value = 1.0
    else:
        value = feature["value_function"](max_value)

    temp = [
        value,
    ]

    return (
        temp,
        util.sse(sim_data_values, exp_data_values),
        len(sim_data_values),
        sim_time_values,
        sim_data_values,
        exp_data_values,
        [1.0 - i for i in temp],
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    temp["value_function"] = score.value_function(feature["max_value"], abstol)
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = ["%s_Height" % name]
    return temp



import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
import numpy
import numpy.linalg
from addict import Dict
import sys
import CADETMatch.smoothing as smoothing
import multiprocessing

name = "DextranSSE"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = False
    settings.badScore = -sys.float_info.max
    settings.meta_mask = True
    settings.count = 1
    return settings


def run(sim_data, feature):
    "special score designed for dextran. This looks at only the front side of the peak up to the maximum slope and pins a value at the elbow in addition to the top"
    exp_time_values = feature["time"]

    min_value = feature["min_value"]
    max_value = feature["max_value"]

    selected = feature["selected"]

    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)

    if max(sim_data_values) < max_value:  # the system has no point higher than the value we are looking for
        # remove hard failure
        max_value = max(sim_data_values)

    exp_time_values = exp_time_values[selected]
    exp_data_zero = feature["exp_data_zero"]

    sim_data_values_smooth = smoothing.smooth_data(
        sim_time_values, sim_data_values, feature["critical_frequency"], feature["smoothing_factor"]
    )
    sim_data_zero, sim_min_time, sim_min_value, sim_max_time, sim_max_value = score.cut_zero(
        sim_time_values, sim_data_values_smooth, min_value, max_value
    )

    sse = util.sse(sim_data_zero, exp_data_zero)

    data = [-sse,], sse, len(sim_data_zero), sim_time_values, sim_data_zero, exp_data_zero, [sse,]

    return data


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    # change the stop point to be where the max positive slope is along the searched interval
    name = "%s_%s" % (sim.root.experiment_name, feature["name"])
    s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(selectedTimes, selectedValues, name, cache)
    smooth_value, values_der = smoothing.full_smooth(selectedTimes, selectedValues, crit_fs, s, crit_fs_der)

    spline_der = scipy.interpolate.InterpolatedUnivariateSpline(selectedTimes, values_der, ext=1)
    spline = scipy.interpolate.InterpolatedUnivariateSpline(selectedTimes, smooth_value, ext=1)

    min_time, min_value, max_time, max_value = score.find_cuts(selectedTimes, smooth_value, spline, spline_der)

    exp_data_zero, exp_min_time, exp_min_value, exp_max_time, exp_max_value = score.cut_zero(
        selectedTimes, smooth_value, min_value, max_value
    )

    if exp_min_time is not None:
        min_time = exp_min_time
        min_value = exp_min_value
        max_time = exp_max_time
        max_value = exp_max_value

    multiprocessing.get_logger().info("Dextran %s  start: %s   stop: %s  max value: %s", name, min_time, max_time, max_value)

    temp["min_time"] = feature["start"]
    temp["max_time"] = feature["stop"]
    temp["min_value"] = min_value
    temp["max_value"] = max_value
    temp["exp_data_zero"] = exp_data_zero
    temp["peak_max"] = max_value
    temp["smoothing_factor"] = s
    temp["critical_frequency"] = crit_fs
    temp["critical_frequency_der"] = crit_fs_der
    temp["smooth_value"] = smooth_value
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = [
        "%s_SSE" % name,
    ]
    return temp



import CADETMatch.util as util
import CADETMatch.evo as evo
import CADETMatch.pareto as pareto
from pathlib import Path
import scipy.optimize
import numpy
import numpy.linalg
import hashlib
import csv
import time

import CADETMatch.cache as cache
import multiprocessing
from cadet import Cadet, H5


class GradientException(Exception):
    pass


def create_template(experiment, name, tolerance, cache, settings, template_name, store_name):
    simulationGrad = Cadet(experiment["simulation"].root)
    template_path_grad = Path(settings["resultsDirMisc"], template_name % name)
    simulationGrad.filename = template_path_grad.as_posix()
        
    if cache.dynamicTolerance:
        simulationGrad.root.input.solver.time_integrator.abstol = tolerance
        simulationGrad.root.input.solver.time_integrator.reltol = 0.0

    start = time.time()
    util.runExperiment(None, experiment, cache.settings, cache.target, simulationGrad, experiment.get("timeout", 1800), cache)
    elapsed = time.time() - start

    # timeout needs to be stored in the template so all processes have it without calculating it
    simulationGrad.root.timeout = max(10, elapsed * 50)
    simulationGrad.save()

    multiprocessing.get_logger().info("grad simulation took %s", elapsed)

    multiprocessing.get_logger().info(
        "grad %s abstol=%.3g  reltol=%.3g",
        simulationGrad.filename,
        simulationGrad.root.input.solver.time_integrator.abstol,
        simulationGrad.root.input.solver.time_integrator.reltol,
    )

    experiment[store_name] = simulationGrad

def setupTemplates(cache):
    settings = cache.settings
    target = cache.target

    for experiment in settings["experiments"]:
        name = experiment["name"]

        abstolFactorGrad = util.get_grad_tolerance(cache, name)

        create_template(experiment, name, abstolFactorGrad, cache, settings, "template_%s_grad.h5", "simulationSens")

def grad_score(values, cache):
    values = numpy.array(values)

    if cache.allScoreSSE:
        values = -values

    prod = scipy.stats.gmean(values)

    if not cache.allScoreSSE:
        prod = 1 - prod
    return prod

def search(
    gradCheck, offspring, cache, writer, csvfile, grad_hof, meta_hof, generation, check_all=False, result_data=None, filterOverlap=True
):
    if check_all:
        checkOffspring = offspring
    else:
        checkOffspring = (ind for ind in offspring if grad_score(ind.fitness.values, cache) <= gradCheck)
    if filterOverlap:
        checkOffspring = filterOverlapArea(cache, checkOffspring)
    newOffspring = cache.toolbox.map(cache.toolbox.evaluate_grad, map(tuple, checkOffspring))

    temp = []
    csv_lines = []
    meta_csv_lines = []

    gradient_results = []

    multiprocessing.get_logger().info("starting coarse refine")
    new_results = processOffspring(
        newOffspring, temp, csv_lines, meta_csv_lines, gradient_results, grad_hof, meta_hof, generation, result_data, cache
    )
    multiprocessing.get_logger().info("ending coarse refine")

    if new_results:
        multiprocessing.get_logger().info("starting fine refine")
        fineOffspring = cache.toolbox.map(cache.toolbox.evaluate_grad_fine, map(tuple, meta_hof))
        processOffspring(
            fineOffspring, temp, csv_lines, meta_csv_lines, gradient_results, grad_hof, meta_hof, generation, result_data, cache
        )
        multiprocessing.get_logger().info("ending fine refine")

    if temp:
        best = meta_hof.getBestScores()
        prod = best[0]
        if cache.allScoreSSE:
            prod = -prod
        else:
            prod = 1 - prod
        if prod < gradCheck:
            gradCheck = prod
            multiprocessing.get_logger().info("updating gradCheck %s", gradCheck)

    writer.writerows(csv_lines)

    # flush before returning
    csvfile.flush()

    path_meta_csv = cache.settings["resultsDirMeta"] / "results.csv"
    with path_meta_csv.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
        writer.writerows(meta_csv_lines)

    if gradient_results:
        write_gradient_results(gradient_results, cache)

    util.cleanupFront(cache, None, meta_hof, grad_hof)
    util.writeMetaFront(cache, meta_hof, path_meta_csv)

    return gradCheck, temp


def processOffspring(offspring, temp, csv_lines, meta_csv_lines, gradient_results, grad_hof, meta_hof, generation, result_data, cache):
    new_meta = []
    for i in offspring:
        if i is None:
            pass
        elif i.x is not None:
            gradient_results.append(i)

            ind = cache.toolbox.individual_guess(i.x)

            fit, csv_line, meta_score, results, individual = cache.toolbox.evaluate_final(ind, run_experiment=runExperimentSens)

            ind.fitness.values = fit

            csv_line[0] = "GRAD"

            try:
                csv_line[1] = numpy.linalg.cond(i.jac)
            except (AttributeError, numpy.linalg.LinAlgError):
                csv_line[1] = ""

            save_name_base = hashlib.md5(str(list(ind)).encode("utf-8", "ignore")).hexdigest()
            ind.csv_line = [time.ctime(), save_name_base] + csv_line

            ind_meta = cache.toolbox.individualMeta(ind)
            ind_meta.fitness.values = meta_score
            ind_meta.csv_line = [time.ctime(), save_name_base] + csv_line

            util.update_result_data(cache, ind, fit, result_data, results, meta_score)

            if csv_line:
                csv_lines.append([time.ctime(), save_name_base] + csv_line)
                onFront, significant = pareto.updateParetoFront(grad_hof, ind, cache)
                if onFront and not cache.metaResultsOnly:
                    util.processResultsGrad(save_name_base, ind, cache, results)

                onFrontMeta, significant = pareto.updateParetoFront(meta_hof, ind_meta, cache)
                new_meta.append(onFrontMeta)
                if onFrontMeta:
                    meta_csv_lines.append([time.ctime(), save_name_base] + csv_line)
                    util.processResultsMeta(save_name_base, ind, cache, results)

                    if significant:
                        cache.lastProgressGeneration = generation

            temp.append(ind)
    return any(new_meta)


def write_gradient_results(gradient_results, cache):
    grad_path = Path(cache.settings["resultsDir"]) / "grad_result.h5"
    results = H5()
    results.filename = grad_path.as_posix()

    if grad_path.exists():
        results.load()

    for i in gradient_results:
        grad_key = " ".join(["%.16f" % val for val in i.cadetValuesInput])

        for key, value in i.items():
            results[grad_key][key] = value
    results.save()


def filterOverlapArea(cache, checkOffspring, cutoff=0.01):
    """if there is no overlap between the simulation and the data there is no gradient to follow and these entries need to be skipped
    This function also sorts from highest to lowest overlap and keeps the top multiStartPercent"""
    checkOffspring = list(checkOffspring)
    temp = cache.toolbox.map(cache.toolbox.evaluate, map(list, checkOffspring))

    temp_offspring = []

    lookup = util.create_lookup(checkOffspring)

    for fit, csv_line, meta_score, results, individual in temp:
        ind = util.pop_lookup(lookup, individual)
        temp_area_total = 0.0
        temp_area_overlap = 0.0
        if results is not None:
            for exp in results.values():
                sim_times = exp["sim_time"]
                sim_values = exp["sim_value"]
                exp_values = exp["exp_value"]

                for sim_time, sim_value, exp_value in zip(sim_times, sim_values, exp_values):
                    if len(sim_time) < len(exp_value) and len(exp_value) % len(sim_time) == 0:
                        # dealing with fractionation data, there are multiple data sets of len(sim_time)
                        exp_value = exp_value.reshape(-1, len(sim_time))
                        sim_value = sim_value.reshape(-1, len(sim_time))
                        for exp_row, sim_row in zip(exp_value, sim_value):
                            temp_area_total += numpy.trapz(exp_row, sim_time)
                            temp_area_overlap += numpy.trapz(numpy.min([sim_row, exp_row], 0), sim_time)
                    else:
                        temp_area_total += numpy.trapz(exp_value, sim_time)
                        temp_area_overlap += numpy.trapz(numpy.min([sim_value, exp_value], 0), sim_time)

            percent = temp_area_overlap / temp_area_total
            if percent > cutoff:
                temp_offspring.append((percent, ind))
                multiprocessing.get_logger().info("kept %s with overlap (%s) in gradient descent", ind, percent)
            else:
                multiprocessing.get_logger().info("removed %s for insufficient overlap (%s) in gradient descent", ind, percent)
        else:
            multiprocessing.get_logger().info("removed %s for failure", ind)

    # sort in descending order, this has the best chance of converging so we can quick abort
    temp_offspring.sort(reverse=True)

    if cache.multiStartPercent < 1.0 and temp_offspring:
        # cut to the top multiStartPercent items with a minimum of 1 item
        temp_offspring = temp_offspring[: max(int(cache.multiStartPercent * len(checkOffspring)), 1)]
        multiprocessing.get_logger().info("gradient overlap cutoff %.2g", temp_offspring[-1][0])

    temp_offspring = [ind for (percent, ind) in temp_offspring]

    if checkOffspring:
        multiprocessing.get_logger().info("overlap okay offspring %s", temp_offspring)

    return temp_offspring


def gradSearchFine(x, json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings["resultsDirLog"], "main.log")
        cache.cache.setup(json_path)
    return refine(x, cache.cache.gradFineStop)


def gradSearch(x, json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings["resultsDirLog"], "main.log")
        cache.cache.setup(json_path)
    return refine(x, 1e-4)

def refine(x, xtol):
    localRefine = cache.cache.settings.get("localRefine", "gradient")
    if localRefine == "gradient":
        try:
            x = numpy.clip(x, cache.cache.MIN_VALUE, cache.cache.MAX_VALUE)
            val = scipy.optimize.least_squares(
                fitness_sens_grad,
                x,
                jac="3-point",
                method="trf",
                bounds=(cache.cache.MIN_VALUE, cache.cache.MAX_VALUE),
                xtol=xtol,
                ftol=None,
                gtol=None,
                loss="linear",
                diff_step=1e-4
            )

            cadetValues, cadetValuesExtended = util.convert_individual(val.x, cache.cache)
            cadetValuesInput, cadetValuesExtendedInput = util.convert_individual(x, cache.cache)

            val.cadetValues = cadetValues
            val.cadetValuesExtended = cadetValuesExtended
            val.cadetValuesInput = cadetValuesInput
            val.cadetValuesExtendedInput = cadetValuesExtendedInput
            val.cond = numpy.linalg.cond(val.jac)

            multiprocessing.get_logger().info("gradient optimization result start: %s (%s) result: %s", x, cadetValues, val)

            return val
        except GradientException:
            # If the gradient fails return None as the point so the optimizer can adapt
            return None

    if localRefine == "powell":

        def goal(x):
            diff = fitness_grad(x)
            return float(numpy.sum(diff ** 2.0))

        try:
            x = numpy.clip(x, cache.cache.MIN_VALUE, cache.cache.MAX_VALUE)
            val = scipy.optimize.minimize(goal, x, method="powell", options={"xtol": xtol}, bounds=list(zip(cache.cache.MIN_VALUE, cache.cache.MAX_VALUE)))

            cadetValues, cadetValuesExtended = util.convert_individual(val.x, cache.cache)
            cadetValuesInput, cadetValuesExtendedInput = util.convert_individual(x, cache.cache)

            val.cadetValues = cadetValues
            val.cadetValuesExtended = cadetValuesExtended
            val.cadetValuesInput = cadetValuesInput
            val.cadetValuesExtendedInput = cadetValuesExtendedInput

            multiprocessing.get_logger().info("gradient optimization result start: %s (%s) result: %s", x, cadetValues, val)

            return val
        except GradientException:
            # If the gradient fails return None as the point so the optimizer can adapt
            return None

def fitness_grad(individual, finished=0):
    return fitness_base(evo.runExperiment, individual, finished)

def fitness_sens_grad(individual, finished=0):
    return fitness_sens(individual, finished)

def fitness(individual, finished=1):
    return fitness_base(evo.runExperiment, individual, finished)

def fitness_sens(individual, finished=1):
    return fitness_base(runExperimentSens, individual, finished)

def fitness_base(fit, individual, finished):
    minimize = []
    error = 0.0

    diff = []

    results = {}
    for experiment in cache.cache.settings["experiments"]:
        setup_experiments(cache.cache, experiment)
        result = fit(individual, "simulationSens", experiment, cache.cache.settings, cache.cache.target, cache.cache)

        if result is not None:
            results[experiment["name"]] = result
            minimize.extend(results[experiment["name"]]["minimize"])
            error += results[experiment["name"]]["error"]
            diff.extend(results[experiment["name"]]["diff"])
        else:
            raise GradientException("Gradient caused simulation failure, aborting")

    # need to minimize
    if cache.cache.gradVector:
        return numpy.array(diff)
    else:
        minimize = numpy.array(minimize)
        if cache.cache.allScoreNorm:
            return minimize
        elif cache.cache.allScoreSSE:
            return numpy.array(diff)
        else:
            multiprocessing.get_logger().info("Norm 1 and SSE based scores can't be mixed %s", cache.cache.badScores)
            sys.exit()

def saveExperimentsSens(save_name_base, settings, target, results):
    return util.saveExperiments(save_name_base, settings, target, results, settings["resultsDirGrad"], "%s_%s_GRAD.h5")

def setup_experiments(cache, experiment):
    if "simulationSens" not in experiment:
        templatePath = Path(cache.settings["resultsDirMisc"], "template_%s_grad.h5" % experiment["name"])
        templateSim = Cadet()
        templateSim.filename = templatePath.as_posix()
        templateSim.load()
        experiment["simulationSens"] = templateSim

def runExperimentSens(individual, template_name, experiment, settings, target, cache):
    setup_experiments(cache, experiment)

    return util.runExperiment(
        individual, experiment, settings, target, experiment[template_name], experiment[template_name].root.timeout, cache
    )



# This module simulates noise (systemic and random) to use for kernel density estimation based on experimental data

# Currently the randomness is

# Pump flow rate
# Pump delay
# Base noise
# Signal noise

import itertools
import CADETMatch.score as score
import CADETMatch.kde_util as kde_util
from cadet import Cadet, H5
from pathlib import Path
import numpy
import multiprocessing
from sklearn.neighbors import KernelDensity
from sklearn import preprocessing
import copy

bw_tol = 1e-4

import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    import h5py

import scipy.optimize

import CADETMatch.util as util
import CADETMatch.cache as cache

import CADETMatch.synthetic_error as synthetic_error

import joblib
import subprocess
import sys

import SALib.sample.sobol_sequence
import scipy.stats

def mirror(data):
    data_max = numpy.max(data, 0)

    mirror_index = data_max <= 1.0
    keep_index = data_max > 1.0

    data_min = data_max - data
    data_mask = numpy.ma.masked_equal(data_min, 0.0, copy=False)
    min_value = data_mask.min(axis=0)

    data_mirror = numpy.zeros(data.shape)

    data_mirror[:, mirror_index] = (
        data_max[mirror_index] + data_max[mirror_index] - numpy.copy(data[:, mirror_index]) + min_value[mirror_index]
    )
    data_mirror[:, keep_index] = data[:, keep_index]
    full_data = numpy.vstack([data_mirror, data])

    return full_data


def setupKDE(cache):
    scores = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])

    scores_mirror = mirror(scores)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    kde = KernelDensity(kernel='gaussian')

    kde, bandwidth, store = kde_util.get_bandwidth(kde, scores_scaler)

    kde.fit(scores_scaler)

    probability = kde.score_samples(scores_scaler)

    joblib.dump(scaler, mcmcDir / "kde_scaler.joblib")

    joblib.dump(kde, mcmcDir / "kde_score.joblib")

    h5_data = H5()
    h5_data.filename = (mcmcDir / "kde_settings.h5").as_posix()
    h5_data.root.bandwidth = bandwidth
    h5_data.root.store = numpy.array(store)
    h5_data.root.scores = scores
    h5_data.root.scores_mirror = scores_mirror
    h5_data.root.scores_mirror_scaled = scores_scaler
    h5_data.root.probability = probability
    h5_data.save()

    cwd = str(Path(__file__).parent)
    ret = subprocess.run(
        [sys.executable, "graph_kde.py", str(cache.json_path), str(util.getCoreCounts())],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
    )
    util.log_subprocess("graph_kde.py", ret)

    return kde, scaler


def getScaler(data):
    scaler = preprocessing.RobustScaler().fit(data)
    return scaler


def getKDE(cache):
    mcmcDir = Path(cache.settings["resultsDirMCMC"])

    kde = joblib.load(mcmcDir / "kde_score.joblib")

    scaler = joblib.load(mcmcDir / "kde_scaler.joblib")

    return kde, scaler


def getKDEPrevious(cache):
    if "mcmc_h5" in cache.settings:
        mcmc_h5 = Path(cache.settings["mcmc_h5"])
        mcmcDir = mcmc_h5.parent

        if mcmcDir.exists():
            kde = joblib.load(mcmcDir / "kde_prior.joblib")

            scaler = joblib.load(mcmcDir / "kde_prior_scaler.joblib")

            return kde, scaler
    return None, None


def generate_data(cache):
    scores = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    save_scores = mcmcDir / "scores_used.npy"

    numpy.save(save_scores, scores)

    scores_mirror = mirror(scores)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    bandwidth, store = get_bandwidth(scores_scaler)

    return scores, bandwidth


def synthetic_error_simulation(x):
    json_path, error = x
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings["resultsDirLog"], "main.log")
        cache.cache.setup(json_path)

    scores = []
    outputs = {}
    simulations = {}
    experiment_failed = False

    delays_store = []
    flows_store = []
    load_store = []
    uv_store = {}
    uv_store_norm = {}

    for experiment in cache.cache.settings["kde_synthetic"]:
        delay_settings = experiment["delay"]
        flow_settings = experiment["flow"]
        load_settings = experiment["load"]
        experimental_csv = experiment["experimental_csv"]
        uv_noise = experiment.get("uv_noise", None)
        uv_noise_norm = experiment.get("uv_noise_norm", None)
        units = experiment["units"]
        name = experiment["name"]

        data = numpy.loadtxt(experimental_csv, delimiter=",")
        times = data[:, 0]

        if "file_path" in experiment:
            template_path = Path(experiment["file_path"])
        else:
            resultsDir = cache.cache.settings["resultsDir"]
            if "resultsDirOriginal" in cache.cache.settings:
                resultsDir = Path(cache.cache.settings["resultsDirOriginal"])

            template_path = resultsDir / "misc" / ("template_%s_base.h5" % name)

        temp = Cadet()
        temp.filename = template_path.as_posix()
        temp.load()

        util.setupSimulation(temp, times, name, cache.cache)

        nsec = temp.root.input.solver.sections.nsec

        for unit in units:
            unit_name = "unit_%03d" % unit
            for comp in range(temp.root.input.model[unit_name].ncomp):
                comp_name = "solution_outlet_comp_%03d" % comp

                if uv_noise is None:
                    error_uv = numpy.zeros(len(temp.root.input.solver.user_solution_times))
                else:
                    error_uv = numpy.random.normal(uv_noise[0], uv_noise[1], len(temp.root.input.solver.user_solution_times))
                uv_store["%s_%s_%s" % (name, unit, comp)] = error_uv

                if uv_noise_norm is None:
                    error_uv = numpy.ones(len(temp.root.input.solver.user_solution_times))
                else:
                    error_uv = numpy.random.normal(uv_noise_norm[0], uv_noise_norm[1], len(temp.root.input.solver.user_solution_times))
                uv_store_norm["%s_%s_%s" % (name, unit, comp)] = error_uv

        def post_function(simulation):
            # baseline drift need to redo this
            # error_slope = numpy.random.normal(error_slope_settings[0], error_slope_settings[1], 1)[0]

            for unit in units:
                unit_name = "unit_%03d" % unit
                for comp in range(simulation.root.input.model[unit_name].ncomp):
                    comp_name = "solution_outlet_comp_%03d" % comp
                    error_uv = uv_store["%s_%s_%s" % (name, unit, comp)]
                    error_norm = uv_store_norm["%s_%s_%s" % (name, unit, comp)]
                    simulation.root.output.solution[unit_name][comp_name] = (
                        simulation.root.output.solution[unit_name][comp_name] * error_norm + error_uv
                    )

        error_delay = Cadet(temp.root)

        #delays = numpy.random.uniform(delay_settings[0], delay_settings[1], nsec)
        delays = error[name]['pump_delays']
        delays_store.extend(delays)

        synthetic_error.pump_delay(error_delay, delays)

        #flow = numpy.random.normal(flow_settings[0], flow_settings[1], error_delay.root.input.solver.sections.nsec)
        flow = error[name]['flow_rates']
        flows_store.extend(flow)

        synthetic_error.error_flow(error_delay, flow)

        #load = numpy.random.normal(load_settings[0], load_settings[1], error_delay.root.input.solver.sections.nsec)
        load = error[name]['loading_concentrations']
        load_store.extend(load)

        synthetic_error.error_load(error_delay, load)

        exp_info = None
        for exp in cache.cache.settings["experiments"]:
            if exp["name"] == name:
                exp_info = exp
                break

        result = util.runExperiment(
            None,
            exp_info,
            cache.cache.settings,
            cache.cache.target,
            error_delay,
            error_delay.root.timeout,
            cache.cache,
            post_function=post_function,
        )

        if result is not None:
            scores.extend(result["scores"])

            simulations[name] = result["simulation"]

            for unit in units:
                unit_name = "unit_%03d" % unit
                for comp in range(result["simulation"].root.input.model[unit_name].ncomp):
                    outputs["%s_unit_%03d_comp_%03d" % (name, int(unit), comp)] = result["simulation"].root.output.solution[
                        "unit_%03d" % int(unit)
                    ]["solution_outlet_comp_%03d" % comp]
        else:
            experiment_failed = True

    if experiment_failed:
        return None, None, None, None

    errors = {}
    errors["delays"] = delays_store
    errors["flows"] = flows_store
    errors["load"] = load_store
    errors["uv_store"] = uv_store

    return scores, simulations, outputs, errors

def get_section_counts(cache):
    sections = []

    for experiment in cache.cache.settings["kde_synthetic"]:
        name = experiment["name"]

        if "file_path" in experiment:
            template_path = Path(experiment["file_path"])
        else:
            resultsDir = cache.cache.settings["resultsDir"]
            if "resultsDirOriginal" in cache.cache.settings:
                resultsDir = Path(cache.cache.settings["resultsDirOriginal"])

            template_path = resultsDir / "misc" / ("template_%s_base.h5" % name)

        temp = Cadet()
        temp.filename = template_path.as_posix()
        temp.load()

        nsec = temp.root.input.solver.sections.nsec

        sections.append(nsec)

    return sections

def generate_error_sequence(cache):
    errors = {}

    for experiment in cache.settings["kde_synthetic"]:
        delay_settings = experiment["delay"]
        flow_settings = experiment["flow"]
        load_settings = experiment["load"]
        count = experiment['count']
        name = experiment["name"]

        if "file_path" in experiment:
            template_path = Path(experiment["file_path"])
        else:
            resultsDir = cache.cache.settings["resultsDir"]
            if "resultsDirOriginal" in cache.settings:
                resultsDir = Path(cache.settings["resultsDirOriginal"])

            template_path = resultsDir / "misc" / ("template_%s_base.h5" % name)

        temp = Cadet()
        temp.filename = template_path.as_posix()
        temp.load()

        nsec = temp.root.input.solver.sections.nsec

        error = {}

        #total space is 1 delay per section 1 flow rate per new section and 1 load per new section
        delay_size = nsec
        flow_size = 2*nsec
        load_size = 2*nsec
        total_size = delay_size + flow_size + load_size

        errors_exp = SALib.sample.sobol_sequence.sample(count, total_size)

        #set delays
        block = errors_exp[:,:delay_size]
        if delay_settings[0] != delay_settings[1]:
            block = scipy.stats.uniform.ppf(block, delay_settings[0], delay_settings[1])
        else:
            block = numpy.ones(block.shape) * delay_settings[0]

        error['pump_delays'] = block

        #set flow range changes
        lb = delay_size
        ub = delay_size + flow_size

        #grab the block to work with easier
        block = errors_exp[:,lb:ub]

        #sobol sequence is [0, 1) and 0 needs to be removed and set to the same distance as the closest distance to 1
        min_value = 1 - numpy.max(block)

        block[block < min_value] = min_value

        if flow_settings[1] != 0:
            block = scipy.stats.norm.ppf(block, flow_settings[0], flow_settings[1])
        else:
            block = numpy.ones(block.shape) * flow_settings[0]

        error['flow_rates'] = block


        #set loading changes
        lb = ub
        ub = lb + load_size

        #grab the block to work with easier
        block = errors_exp[:,lb:ub]

        #sobol sequence is [0, 1) and 0 needs to be removed and set to the same distance as the closest distance to 1
        min_value = 1 - numpy.max(block)

        block[block < min_value] = min_value        

        if load_settings[1] != 0:
            block = scipy.stats.norm.ppf(block, load_settings[0], load_settings[1])
        else:
            block = numpy.ones(block.shape) * load_settings[0]

        error['loading_concentrations'] = block

        errors[name] = error

    return errors

def split_errors(errors_all):
    errors = []
    for name,experiment in errors_all.items():
        length = len(experiment['flow_rates'])
        break

    temp = {}
    for key in errors_all:
        temp[key] = {}

    for i in range(length):
        errors.append({key:{} for key in errors_all})
    
    for name,experiment in errors_all.items():
        for error_name, error_value in experiment.items():
            for idx in range(length):
                errors[idx][name][error_name] = error_value[idx,:]

    return errors

def keep_data(data, lb=0.5, ub=99.5):
    cutoffs_lb, cutoffs_ub = numpy.percentile(data, [lb, ub], axis=0)
    remove = numpy.zeros(data.shape)
    
    for i in range(data.shape[1]-2):
        remove[:,i] = data[:,i] > cutoffs_lb[i]
    remove[:,-2] = (data[:,-2] > cutoffs_lb[-2]) & (data[:,-2] < cutoffs_ub[-2])
    remove[:,-1] = (data[:,-1] > cutoffs_lb[-1]) & (data[:,-1] < cutoffs_ub[-1])
    
    remove = numpy.all(remove, axis=1)
    
    return remove

def generate_synthetic_error(cache):
    if "kde_synthetic" in cache.settings:
        count_settings = int(cache.settings["kde_synthetic"][0]["count"])

        scores_all = []
        times = {}
        outputs_all = {}

        uv_store_all = {}

        errors_all = generate_error_sequence(cache)

        errors_split = split_errors(errors_all)
        json_path = itertools.repeat(cache.json_path)

        indexes = []

        for idx, (scores, simulations, outputs, errors) in enumerate(cache.toolbox.map(synthetic_error_simulation, zip(json_path, errors_split))):
            if scores and simulations and outputs:
                indexes.append(idx)

                for key, value in errors["uv_store"].items():
                    uv = uv_store_all.get(key, [])
                    uv.append(value)
                    uv_store_all[key] = uv

                scores_all.append(scores)

                for key, value in outputs.items():
                    temp = outputs_all.get(key, [])
                    temp.append(value)
                    outputs_all[key] = temp

                for key, sim in simulations.items():
                    if key not in times:
                        times[key] = sim.root.output.solution.solution_times

        indexes = numpy.array(indexes)

        for name,experiment in errors_all.items():
            for error_name, error_value in experiment.items():
                errors_all[name][error_name] = error_value[indexes,:]

        scores = numpy.array(scores_all)

        keep_idx = keep_data(scores)

        kept = int(numpy.sum(keep_idx))
        removed = int(len(scores) - kept)

        dir_base = cache.settings.get("resultsDirBase")
        file = dir_base / "kde_data.h5"

        kde_data = H5()
        kde_data.filename = file.as_posix()

        kde_data.root.kept = kept
        kde_data.root.removed = removed

        kde_data.root.scores = scores[keep_idx, :]

        kde_data.root.original.scores = scores

        for output_name, output in outputs_all.items():
            kde_data.root[output_name] = numpy.array(output)[keep_idx,:]
            kde_data.root.original[output_name] = numpy.array(output)[keep_idx,:]

        for time_name, time in times.items():
            kde_data.root["%s_time" % time_name] = time
            kde_data.root.original["%s_time" % time_name] = time

        for name,experiment in errors_all.items():
            for error_name, error_value in experiment.items():
                kde_data.root.errors[name][error_name] = error_value[keep_idx, :]
                kde_data.root.original.errors[name][error_name] = error_value[keep_idx, :]

        for key, value in uv_store_all.items():
            kde_data.root.uv_store[key] = numpy.array(value)[keep_idx,:]
            kde_data.root.original.uv_store[key] = numpy.array(value)[keep_idx,:]

        kde_data.save()

        return scores[keep_idx,:]

    return None


def convert_to_array(seq):
    "this converts a sequence of arrays to a single numpy array and pads uneven rows with 0"
    max_len = numpy.max([len(a) for a in seq])
    return numpy.asarray([numpy.pad(a, (0, max_len - len(a)), "constant", constant_values=0) for a in seq])



import numpy
import numpy.linalg
import scipy
import scipy.signal
import CADETMatch.util as util
import warnings

import multiprocessing

from cadet import H5

butter_order = 3


def get_p(x, y):
    x = numpy.array(x)
    y = numpy.array(y)

    sort_idx = numpy.argsort(x)

    x = x[sort_idx]
    y = y[sort_idx]

    y_min = y - min(y)
    x_min = x - min(x)

    p3 = numpy.array([x_min, y_min]).T
    factor = numpy.max(p3, 0)
    p3 = p3 / factor
    p1 = p3[0, :]
    p2 = p3[-1, :]

    return x, min(x), y, min(y), p1, p2, p3, factor


def refine_butter(times, values, x, y, fs, start):
    x, x_min, y, y_min, p1, p2, p3, factor = get_p(x, y)

    def goal(crit_fs):
        crit_fs = 10.0 ** crit_fs[0]
        try:
            sos = scipy.signal.butter(butter_order, crit_fs, btype="lowpass", analog=False, fs=fs, output="sos")
        except ValueError:
            return 1e6
        low_passed = scipy.signal.sosfiltfilt(sos, values)

        sse = numpy.sum((low_passed - values) ** 2)

        pT = numpy.array([crit_fs - x_min, numpy.log(sse) - y_min]).T / factor

        d = numpy.cross(p2 - p1, p1 - pT) / numpy.linalg.norm(p2 - p1)

        return -d

    lb = numpy.log10(x[0])
    ub = numpy.log10(x[-1])
    
    crit_fs_sample = numpy.linspace(lb, ub, 50)
    errors = numpy.array([goal([i]) for i in crit_fs_sample])
    
    root, fs_x, fs_y = util.find_opt_poly(crit_fs_sample, errors, numpy.argmin(errors))
    
    result = scipy.optimize.minimize(goal, root, method="powell", bounds=[(fs_x[0], fs_x[-1]),])
    crit_fs = 10 ** result.x[0]

    return crit_fs


def refine_smooth(times, values, x, y, start, name):
    x, x_min, y, y_min, p1, p2, p3, factor = get_p(x, y)
    if name is None:
        name = "unknown"

    def goal(s):
        s = 10 ** s[0]
        with warnings.catch_warnings():
            warnings.filterwarnings("error")

            try:
                spline = scipy.interpolate.UnivariateSpline(times, values, s=s, k=5, ext=3)
            except Warning:
                multiprocessing.get_logger().info("caught a warning for %s %s", name, s)
                return 1e6

        pT = numpy.array([s - x_min, len(spline.get_knots()) - y_min]).T / factor
        d = numpy.cross(p2 - p1, p1 - pT) / numpy.linalg.norm(p2 - p1)

        return -d

    lb = numpy.log10(x[0])
    ub = numpy.log10(x[-1])
    
    smoothing_sample = numpy.linspace(lb, ub, 50)
    errors = numpy.array([goal([i]) for i in smoothing_sample])
    
    root, fs_x, fs_y = util.find_opt_poly(smoothing_sample, errors, numpy.argmin(errors))
    
    result = scipy.optimize.minimize(goal, root, method="powell", bounds=[(fs_x[0], fs_x[-1]),])
    s = 10 ** result.x[0]

    spline = scipy.interpolate.UnivariateSpline(times, values, s=s, k=5, ext=3)

    return s, len(spline.get_knots())


def find_L(x, y):
    # find the largest value greater than 0, otherwise return none to just turn off butter filter
    x, x_min, y, y_min, p1, p2, p3, factor = get_p(x, y)

    d = numpy.cross(p2 - p1, p1 - p3) / numpy.linalg.norm(p2 - p1)

    max_idx = numpy.argmax(d)
    max_d = d[max_idx]
    l_x = x[max_idx]
    l_y = y[max_idx]

    if max_d <= 0:
        return None, None

    return l_x, l_y


def find_butter(times, values):
    filters = []
    sse = []

    fs = 1.0 / (times[1] - times[0])

    ub = fs / 2.0
    ub_l = numpy.log10(ub)

    for i in numpy.logspace(-6, ub_l, 50):
        try:
            sos = scipy.signal.butter(butter_order, i, btype="lowpass", analog=False, fs=fs, output="sos")
            low_passed = scipy.signal.sosfiltfilt(sos, values)

            filters.append(i)
            sse.append(numpy.sum((low_passed - values) ** 2))
        except ValueError:
            continue

    L_x, L_y = find_L(filters, numpy.log(sse))

    if L_x is not None:
        L_x = refine_butter(times, values, filters, numpy.log(sse), fs, L_x)

    return L_x


def smoothing_filter_butter(times, values, crit_fs):
    if crit_fs is None:
        return values
    fs = 1.0 / (times[1] - times[0])

    sos = scipy.signal.butter(butter_order, crit_fs, btype="lowpass", analog=False, fs=fs, output="sos")
    low_passed = scipy.signal.sosfiltfilt(sos, values)
    return low_passed


def load_data(name, cache):
    crit_fs = None
    crit_fs_der = None
    s = None
    s_knots = 0

    # quick abort
    if name is None or cache is None:
        return s, crit_fs, crit_fs_der

    factor_file = cache.settings["resultsDirMisc"] / "find_smoothing_factor.h5"

    data = H5()
    data.filename = factor_file.as_posix()

    if factor_file.exists():
        data.load()

    if name in data.root:
        s = float(data.root[name].s)

        crit_fs = data.root[name].crit_fs
        if crit_fs == -1.0:
            crit_fs = None

        crit_fs_der = data.root[name].crit_fs_der
        if crit_fs_der == -1.0:
            crit_fs_der = None

        s_knots = int(data.root[name].s_knots)
    else:
        return s, crit_fs, crit_fs_der

    crit_fs_message = "critical frequency disable"
    if crit_fs is not None:
        crit_fs_message = "critical frequency %.3e" % crit_fs

    crit_fs_der_message = "critical frequency der disable"
    if crit_fs_der is not None:
        crit_fs_der_message = "critical frequency der %.3e" % crit_fs_der

    multiprocessing.get_logger().info("smoothing_factor %s  %.3e  %s  %s knots %d", name, s, crit_fs_message, crit_fs_der_message, s_knots)

    return s, crit_fs, crit_fs_der


def find_smoothing_factors(times, values, name, cache):
    times, values = resample(times, values)
    min = 1e-2

    s, crit_fs, crit_fs_der = load_data(name, cache)

    if s is not None:
        return s, crit_fs, crit_fs_der

    # normalize the data
    values = values * 1.0 / max(values)

    crit_fs = find_butter(times, values)

    if crit_fs is None:
        multiprocessing.get_logger().info("%s butter filter disabled, no viable L point found", name)

    values_filter = smoothing_filter_butter(times, values, crit_fs)

    spline = scipy.interpolate.UnivariateSpline(times, values_filter, s=min, k=5, ext=3)
    knots = []
    all_s = []

    knots.append(len(spline.get_knots()))
    all_s.append(min)

    # This limits to 1e-14 max smoothness which is way beyond anything normal
    for i in range(1, 200):
        s = min / (1.1 ** i)
        with warnings.catch_warnings():
            warnings.filterwarnings("error")

            try:
                spline = scipy.interpolate.UnivariateSpline(times, values_filter, s=s, k=5, ext=3)
                knots.append(len(spline.get_knots()))
                all_s.append(s)

                if len(spline.get_knots()) > 600:
                    break

            except Warning:
                multiprocessing.get_logger().info("caught a warning for %s %s", name, s)
                break

    knots = numpy.array(knots)
    all_s = numpy.array(all_s)

    s, s_knots = find_L(all_s, knots)

    if s is not None:
        s, s_knots = refine_smooth(times, values_filter, all_s, knots, s, name)

    spline, factor = create_spline(times, values, crit_fs, s)

    # run a quick butter pass to remove high frequency noise in the derivative (needed for some experimental data)
    values_filter = spline.derivative()(times) / factor
    crit_fs_der = find_butter(times, values_filter)

    record_smoothing(s, s_knots, crit_fs, crit_fs_der, knots, all_s, name, cache)

    return s, crit_fs, crit_fs_der


def record_smoothing(s, s_knots, crit_fs, crit_fs_der, knots, all_s, name=None, cache=None):
    if name is None or cache is None:
        return
    factor_file = cache.settings["resultsDirMisc"] / "find_smoothing_factor.h5"

    data = H5()
    data.filename = factor_file.as_posix()

    if factor_file.exists():
        data.load()

    if name not in data.root:
        data.root[name].knots = knots
        data.root[name].all_s = all_s
        data.root[name].s = float(s)
        data.root[name].s_knots = int(s_knots)
        if crit_fs is None:
            data.root[name].crit_fs = -1.0
        else:
            data.root[name].crit_fs = float(crit_fs)
        if crit_fs_der is None:
            data.root[name].crit_fs_der = -1.0
        else:
            data.root[name].crit_fs_der = float(crit_fs)
        data.save()

    crit_fs_message = "critical frequency disable"
    if crit_fs is not None:
        crit_fs_message = "critical frequency %.3e" % crit_fs

    crit_fs_der_message = "critical frequency der disable"
    if crit_fs_der is not None:
        crit_fs_der_message = "critical frequency der %.3e" % crit_fs_der

    multiprocessing.get_logger().info("smoothing_factor %s  %.3e  %s  %s knots %d", name, s, crit_fs_message, crit_fs_der_message, s_knots)


def create_spline(times, values, crit_fs, s):
    times, values = resample(times, values)
    factor = 1.0 / max(values)
    values = values * factor
    values_filter = smoothing_filter_butter(times, values, crit_fs)

    return scipy.interpolate.UnivariateSpline(times, values_filter, s=s, k=5, ext=3), factor


def smooth_data(times, values, crit_fs, s):
    spline, factor = create_spline(times, values, crit_fs, s)

    return spline(times) / factor


def smooth_data_derivative(times, values, crit_fs, s, crit_fs_der, smooth=True):
    times_resample, values_resample = resample(times, values)
    spline, factor = create_spline(times_resample, values_resample, crit_fs, s)

    if smooth:
        values_filter_der = spline.derivative()(times_resample) / factor
        factor_der = numpy.max(values_filter_der)
        values_filter_der = butter(times_resample, values_filter_der / factor_der, crit_fs_der) * factor_der
        spline_der = scipy.interpolate.InterpolatedUnivariateSpline(times_resample, values_filter_der, k=5, ext=3)
        values_filter_der = spline_der(times)
    else:
        values_filter_der = spline.derivative()(times) / factor
    return values_filter_der


def full_smooth(times, values, crit_fs, s, crit_fs_der, smooth=True):
    # return smooth data derivative of data
    times_resample, values_resample = resample(times, values)
    spline, factor = create_spline(times_resample, values_resample, crit_fs, s)

    values_filter = spline(times) / factor

    # run a quick butter pass to remove high frequency noise in the derivative (needed for some experimental data)

    if smooth:
        values_filter_der = spline.derivative()(times_resample) / factor
        factor_der = numpy.max(values_filter_der)
        values_filter_der = butter(times_resample, values_filter_der / factor_der, crit_fs_der) * factor_der
        spline_der = scipy.interpolate.InterpolatedUnivariateSpline(times_resample, values_filter_der, k=5, ext=3)
        values_filter_der = spline_der(times)
    else:
        values_filter_der = spline.derivative()(times) / factor
    return values_filter, values_filter_der


def butter(times, values, crit_fs_der):
    factor = 1.0 / max(values)
    values = values * factor

    values_filter = smoothing_filter_butter(times, values, crit_fs_der) / factor

    return values_filter


def resample(times, values):
    diff_times = times[1:] - times[:-1]
    max_time = numpy.max(diff_times)
    min_time = numpy.min(diff_times)
    per = (max_time - min_time) / min_time

    if per > 0.01:
        # time step is not consistent, resample the time steps to a uniform grid based on the smallest time step size seen
        times_resample = numpy.arange(times[0], times[-1], min_time)
        times_resample[-1] = times[-1]
        diff_times = times_resample[1:] - times_resample[:-1]
        max_time = numpy.max(diff_times)
        min_time = numpy.min(diff_times)
        per = (max_time - min_time) / min_time

        spline_resample = scipy.interpolate.InterpolatedUnivariateSpline(times, values, k=5, ext=3)
        values_resample = spline_resample(times_resample)

        return times_resample, values_resample
    else:
        return times, values



import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
import scipy.interpolate
import numpy
from addict import Dict
import CADETMatch.smoothing as smoothing

name = "Shape"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True

    derivative = feature.get("derivative", 1)

    if derivative:
        settings.count = 6
    else:
        settings.count = 3
    return settings


def run(sim_data, feature):
    "similarity, value, start stop"
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)
    selected = feature["selected"]

    exp_data_values = feature["value"][selected]
    exp_time_values = feature["time"][selected]
    exp_data_values_spline = feature["exp_data_values_spline"]

    sim_data_values_smooth, sim_data_values_der_smooth = smoothing.full_smooth(
        exp_time_values, sim_data_values, feature["critical_frequency"], feature["smoothing_factor"], feature["critical_frequency_der"]
    )

    [high, low] = util.find_peak(exp_time_values, sim_data_values_smooth)

    time_high, value_high = high

    pearson, diff_time = score.pearson_spline(exp_time_values, sim_data_values_smooth, feature["smooth_value"])

    derivative = feature.get("derivative", 1)

    if derivative:
        pearson_der, diff_time_der = score.pearson_spline(exp_time_values, sim_data_values_der_smooth, exp_data_values_spline)
        [highs_der, lows_der] = util.find_peak(exp_time_values, sim_data_values_der_smooth)

    temp = [pearson, feature["value_function"](value_high), feature["time_function"](numpy.abs(diff_time))]

    if derivative:
        temp.extend(
            [pearson_der, feature["value_function_high"](highs_der[1]), feature["value_function_low"](lows_der[1]),]
        )

    return (
        temp,
        util.sse(sim_data_values, exp_data_values),
        len(sim_data_values),
        sim_time_values,
        sim_data_values,
        exp_data_values,
        [1.0 - i for i in temp],
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    name = "%s_%s" % (sim.root.experiment_name, feature["name"])
    s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(selectedTimes, selectedValues, name, cache)

    exp_data_values_smooth, exp_data_values_der_smooth = smoothing.full_smooth(selectedTimes, selectedValues, crit_fs, s, crit_fs_der)

    [high, low] = util.find_peak(selectedTimes, exp_data_values_der_smooth)

    temp = {}
    temp["peak"] = util.find_peak(selectedTimes, exp_data_values_smooth)[0]

    decay = feature.get("decay", 0)

    if decay:
        temp["time_function"] = score.time_function_decay(feature["time"][-1])
    else:
        temp["time_function"] = score.time_function(feature["time"][-1], 0.1 * CV_time)

    temp["value_function"] = score.value_function(temp["peak"][1], abstol)
    temp["value_function_high"] = score.value_function(high[1], numpy.abs(high[1]) / 1000)
    temp["value_function_low"] = score.value_function(low[1], numpy.abs(low[1]) / 1000)
    temp["peak_max"] = max(selectedValues)
    temp["smoothing_factor"] = s
    temp["critical_frequency"] = crit_fs
    temp["critical_frequency_der"] = crit_fs_der
    temp["smooth_value"] = exp_data_values_smooth
    temp["exp_data_values_spline"] = exp_data_values_der_smooth
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    derivative = feature.get("derivative", 1)
    temp = ["%s_Similarity" % name, "%s_Value" % name, "%s_Time" % name]

    if derivative:
        temp.extend(["%s_Derivative_Similarity" % name, "%s_Der_High_Value" % name, "%s_Der_Low_Value" % name])
    return temp
